/***************************************************************************
                          CImageAttributes.h  -  description
                             -------------------
    begin                : 01.01.2001
    copyright            : (C) 2001 by Marco Budde
    email                : tuvisionlib@tuvision.de
    origin               : Diplomarbeit Marco Budde

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#ifndef __TUVISION_CIMAGEATTRIBUTES_H__
#define __TUVISION_CIMAGEATTRIBUTES_H__

#include "TUVisionDef.h"
#include <string>

namespace TUVision {

/**********************************************************/
//! Represents an extensible attribute map.
/*! 
 *  Used by the CImage class and its clients to store
 *  attributes like "xdpi" and "filename" and so on 
 *  flexibly.
 *
 * 
 *  A short example:
 *  
 * \code
 * CImageAttributes attribs; // Constructs an empty map.
 *
 * attribs.setDoubleAttribute("xdpi",150.0);
 * attribs.setStringAttribute("filename","My_FileName.tga");
 *
 * ...
 *
 * double xdpi=attribs.getDoubleAttribute("xdpi");
 * string filename=attribs.getStringAttribute("filename");
 *
 * ...
 *
 * // Unknown attributes are automatically defined.
 * attribs.getDoubleAttribute("gamma")=1.7; 
 * // You may even modify the attributes!
 * attribs.getStringAttribute("filename")="new_filename"; 
 *
 * \endcode
 *
 * \warning Due to an outstanding discussion, the feature of 
 *          on-the-fly attrib modification may be removed
 *          (see last code example).
 *
 * \par Commonly used attributes:
 * All attributes are written lowercase, words combined 
 * (in the hope that no one will need too long attribute names)
 * 
 * \par 
 * The attributes that are currently used are:
 * 
 * \par Double Attributes:
 * xdpi - the horizontal resolution (in dots per inch)<br>
 * ydpi - the vertical resolution (in dots per inch)<br>
 * framenumber - the image's position in the stream of images<br>
 * 
 * \par String Attributes:
 * description - the (optional) description of the image
 *
 * \author Eric Sommerlade
 */
/**********************************************************/

class _CCTUVISION CImageAttributes {
public:

    // CONSTRUCTORS

    /******************************************************/
    //! Constructs empty attribute maps. 
    /******************************************************/
    CImageAttributes();

    /******************************************************/
    //! Copies attribute maps from input. 
    /******************************************************/
    CImageAttributes(const CImageAttributes&);

    /******************************************************/
    //! Deletes attribute maps. 
    /******************************************************/
    ~CImageAttributes();


    // GET-/SETTERS

    /******************************************************/
    //! Gets a string attribute.
    /******************************************************/ 
    std::string& getStringAttribute (const std::string& name) const;  

    /******************************************************/ 
    //! Sets a string attribute.
    /******************************************************/ 
    void setStringAttribute (const std::string& name, 
        const std::string& value); 

    /******************************************************/
    //! Gets a double attribute.
    /******************************************************/ 
    double& getDoubleAttribute (const std::string& name) const;   

    /******************************************************/
    //! Sets a double attribute.
    /******************************************************/ 
    void setDoubleAttribute (const std::string& name, 
        double value); 

private:

    class DoubleMap;
    class StringMap;

    //! here the double attributes are stored
    DoubleMap* d_doubles; 

    //! here the string attributes are stored
    StringMap* d_strings; 

    //! protect attributes from assignment
    CImageAttributes& operator=(const CImageAttributes&);
};
} // namespace TUVision

#endif // __TUVISION_CIMAGEATTRIBUTES_H__
