/***************************************************************************
                         CImageType.h  -  description
                            -------------------
    begin                : 13.11.2002
    copyright            : (C) 2001 by Tobias Frommfeld
                           based on Diplomarbeit Marco Budde,
                           Eric Sommerlade, PhD Shdaifat
    email                : tuvisionlib@tuvision.de
    origin               : Studienarbeit Tobias Frommfeld

***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#ifndef __TUVISION_TYPES_H__
#define __TUVISION_TYPES_H__ 1

#include "vigra/numerictraits.hxx"
#include "vigra/imageiterator.hxx" 
#include "vigra/tinyvector.hxx" 

#include "tuv/CImage.h"
#include "CExceptionRange.h"

namespace TUVision {

template <class T> class CImageData;

/**********************************************************/
//!  Template-class defining generic image in the memory.
/*! 
 *   This class defines an generic image. From this class we
 *   derive CImageTemplate. Based on Diplomarbeit Marco Budde,
 *   Eric Sommerlade, PhD Shdaifat
 *
 *   \ingroup image
 *
 *   \author Tobias Frommfeld
 */
/**********************************************************/
template< class TYPE >
class _CCTUVISION CImageType : public CImage {

public:

    typedef	TYPE				value_type;

    /**********************************************************/
    /*! 
     *   Writes the value to the specified position in
     *   the image. This function is a little bit slower than
     *   a direct access to the image buffer, but it makes
     *   the code more readable. Normally it is better to use
     *   getRow() instead.
     *
     *   \param x x coordinate
     *   \param y y coordinate
     *   \param value value to be written (0-255)
     *   \param channel color channel
     */
    /**********************************************************/
    inline void setPixel (int x, int y, TYPE value, int channel = 0) {
        d_img_pnt [(x+y * d_width) * d_nchannels + channel] = value;
    }

    /**********************************************************/
    /*! 
     *   Reads the value of the specified pixel. This function
     *   is a little bit slower than a direct access to the 
     *   image buffer, but it makes the code more readable.
     *   Normally it is better to use getRow() instead.
     *
     *   \param x x coordinate
     *   \param y y coordinate
     *   \param channel color channel
     *   \return the value of the pixel
     */
    /**********************************************************/
    inline TYPE getPixel (int x, int y, int channel = 0) const
    {
        return (d_img_pnt [(x+y * d_width) * d_nchannels + channel]); 
    }

    /**********************************************************/
    /*! 
     *   Get a pointer to the specified row of the image.
     *
     *   \param row the specified row
     */
    /*********************************************************/
    inline TYPE* getRow (int row)
    {
        // Does this row exist?
        if ((row < 0) || (row >= d_height)) 
            throw CExceptionRange ("Row does not exist.",
            row, 0, d_height - 1);

        return (d_img_pnt + (row * d_width) * d_nchannels);
    }

    /**********************************************************/
    /*! 
     *   Get a pointer to the specified row of the image.
     *
     *   \param row the specified row
     */
    /*********************************************************/
    inline const TYPE* getRow (int row) const
    {
        // Does this row exist?
        if ((row < 0) || (row >= d_height)) 
            throw CExceptionRange ("Row does not exist.", 
            row, 0, d_height - 1);

        return (d_img_pnt + (row * d_width) * d_nchannels);
    }

    /**********************************************************/
    /*! 
     *   Sets all pixels (and all of their color channels) in 
     *   the active region to value "0".
     */
    /**********************************************************/
    inline  void clearImage()
    {
        int j;
        for(j = 0; j < getH() ; ++j)
            memset (getRow(j), 0, getW() * d_nchannels * 
            sizeof (TYPE));
    }


protected:

    /**********************************************************/
    /*!
     *   Constructor: create new image object with the
     *   specified dimensions.
     *
     *  \param w          image width
     *  \param h          image height
     *  \param nchannels  number of color channels
     */
    /**********************************************************/

    CImageType (int w, int h, int nchannels)
        : CImage (w, h, nchannels)
    {
        d_img_data = new CImageData<TYPE>(w * h * nchannels);
        resetWindow();
    }


    /**********************************************************/
    /*!
     *   Constructor: copy an image to the local object.
     *   It is possible to share the image buffer memory
     *   between objects.
     *  
     *   If the buffer is shared, the whole buffer of the 
     *   original image is shared, i.e. resetting the window 
     *   size results in access to the full image buffer for 
     *   the specified object.
     *
     *   If the buffer is copied, only the active image region 
     *   is copied. I.e. resetting the window size might have 
     *   no effect.
     *
     *   \param img   image to be copied
     *   \param copy  If true, then copy the image buffer 
     *                between the image objects.
     *                If false, the buffer is shared.
     */
    /**********************************************************/
    CImageType (const CImageType &img, bool copy)
        : CImage (img,copy)
    {  
        const TYPE *src;
        TYPE       *dst;
        int         y;


        if (copy == true) {
            // Yes, we want to copy the image, producing a possibly
            // Smaller version - by using the window of the original image.

            d_img_data = new CImageData<TYPE>
                (d_width * d_height * d_nchannels);

            for (y = 0; y < img.getH(); y++) {
                src = img.getRow (y);
                dst = (TYPE*)(*d_img_data) + d_width * 
                    d_nchannels * y;

                memcpy (dst, src, d_width * d_nchannels * 
                    sizeof (TYPE));
            }

            // Now, adjust to the size of the copy.
            resetWindow();
        }
        else {
            // No, just copy the reference to the image.
            // Now also the window of the original image has to be copied.

            d_img_data = new CImageData<TYPE>(*img.d_img_data);

            // Now, adjust to the windowing of the original image.
            setWindow (img.d_win_x, img.d_win_y, img.d_win_width, 
                img.d_win_height);
        }
    }


    inline virtual void adjustRowPtr()
    {
        d_img_pnt = (TYPE*) (*d_img_data) + 
            d_nchannels * (d_win_y * d_width + d_win_x);
    }

    /**********************************************************/
    /*! 
     *   Inserts an existing image in this image object.
     *   The specified image is copied into this image.
     *
     *   \param img source image
     *   \param x x coordinate of start point
     *   \param y y coordinate of start point
     */
    /**********************************************************/

    inline void insertSubImage (const CImageType &img, int x, int y)
    {
        int                  x_cnt, y_cnt;
        const TYPE			 *pnt_s;
        TYPE		        *pnt_t;


        // test for correct parameters
        if ((x < 0) || (x >= d_width))
            throw CExceptionRange ("You can not insert an image using "
            "this start point (x dimension).", 
            x, 0, d_width - 1);
        if ((y < 0) || (y >= d_height))
            throw CExceptionRange ("You can not insert an image using "
            "this start point (y dimension).", 
            y, 0, d_height - 1);
        if (x + img.getW() > d_width)
            throw CExceptionRange ("The width of the image to be "
            "inserted is to large.",
            img.getW(), 0, d_width - x);
        if (y + img.getH() > d_height)
            throw CExceptionRange ("The height of the image to be "
            "inserted is to large.",
            img.getH(), 0, d_height - y);


        // insert image
        for (y_cnt = 0; y_cnt < img.getH(); y_cnt++)
        {
            pnt_s = img.getRow (y_cnt); 
            pnt_t = getRow (y_cnt + y);

            for (x_cnt = 0; x_cnt < img.getW() * d_nchannels; x_cnt++)
                pnt_t [x_cnt + x*d_nchannels] = pnt_s [x_cnt];
        }
    }

    /**********************************************************/
    /*! 
     *   Check if this image is equal to another image.
     *   The dimensions and all values of the image are
     *   compared.
     *
     *   \param  img the image to be compared with
     *   \return "true" if the images are equal, "false" if 
     *           the images are not equal
     */
    /**********************************************************/
    inline bool isEqual (const CImageType &img) const
    {
        int                  x_cnt, y_cnt;
        const TYPE  *pnt_s;
        const TYPE  *pnt_t;

        // compare the dimensions
        if (getW() != img.getW())
            return (false);

        if (getH() != img.getH())
            return (false);

        // compare the values of the pixels
        for (y_cnt = 0; y_cnt < getH(); y_cnt++)
        {
            pnt_s = img.getRow (y_cnt);
            pnt_t = getRow (y_cnt);

            for (x_cnt = 0; x_cnt < getW() * d_nchannels; x_cnt++)
                if (pnt_t [x_cnt] != pnt_s [x_cnt])
                    return (false);
        }

        return (true);
    }

    /**********************************************************/
    /*! 
     *   Check if this image is not equal to another image.
     *   The dimensions and all values of the image are
     *   compared.
     *
     *   \param img the image to be compared with
     *
     *   \return "false" if the images are equal, "true" if 
     *                       the image are not equal
     */
    /**********************************************************/
    inline bool isUnequal (const CImageType &img) const
    {
        int                  x_cnt, y_cnt;
        const TYPE  *pnt_s;
        const TYPE  *pnt_t;


        // compare the dimensions 
        if (getW() != img.getW())
            return (true);

        if (getH() != img.getH())
            return (true);

        // compare the values of the pixels
        for (y_cnt = 0; y_cnt < getH(); y_cnt++)
        {
            pnt_s = img.getRow (y_cnt);
            pnt_t = getRow (y_cnt);

            for (x_cnt = 0; x_cnt < getW() * d_nchannels; x_cnt++)
                if (pnt_t [x_cnt] != pnt_s [x_cnt])
                    return (true);
        }

        return (false);
    }

private:
    /******************************************************/
    //! Default copy constructor (disabled).
    /******************************************************/
    CImageType (const CImageType &img);

protected:
    CImageData<TYPE>* d_img_data; //!< reference to data
    TYPE *d_img_pnt; //!< pointer to the first available pixel
};


} // namespace TUVision

#endif //  __TUVISION_CIMAGETYPE_H__
