/***************************************************************************
                          CFilterGauss.h  -  description
                             -------------------
    begin                : |19-4-2001|
    copyright            : (C) 2001 by Islam Shdaifat
    email                : shdaifat@tuhh.de
    origin               : PhD Thesis Islam Shdaifat
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#ifndef __TUVISION_FILTERCGAUSS_H__
#define __TUVISION_FILTERCGAUSS_H__

#include "TUVisionDef.h"
#include "tuv/CImageDatatypes.h"

namespace TUVision
{

//! Fast gaussian smoothing filter of the CImageByteGray type.
/*! 
 *  Gaussian Filter.
 *  With the width of the window the filter generates a window of size
 *  2w+1 (alwasy odd). With sigma the standard deviation we generate a
 *  one dimensional matrix cotains a gausian mask. The gaussian filter
 *  is applied to the two dimensional image but it is seperable.
 *  
 *
 *  \code
 *  // create a gausian filter with sigma 6.0 and window size 9 (2*4+1)
 *  TUVision::CFilterGauss gaussFilter(6,4);
 *  CImageByteGray *image;
 *	// Read image from file
 *	gaussFilter.smoothImage(image);
 *	// Save or disply image
 *  \endcode
 *
 *  \author Islam Shdaifat
 *  \ingroup filter
 */ 

class _CCTUVISION CFilterGauss  
{
public:
    CFilterGauss(double sigma = 1, int w = -1);
    virtual ~CFilterGauss();

    //! return the width of the gausian mask
    int getMaskWidth(){ return d_MaskWidth;};

    //! set the width of the gausian mask to new width
    void setMaskWidth(int width);

    //! return the gaussian array
    double *getGaussArray(){ return d_GaussArray;};

    //! set the gaussian array to a new one
    void setGaussArray(double *array);

    //! create the gaussian array from sigam and w
    void createGaussArray(double sigma = 1, int w = 1);

    //! filter the image using the gausian
    void filter(CImageByteGray &image);

    //! filter the image using the gausian
    void filter(CImageByteRGB &image);

    //! return sigma of the mask
    double getSigma(){ return d_sigma;}

    //! set new sigma of the mask
    void setSigma();

    //! return the calue of the gauss curve with sigma at x
    double getGauss(double x, double sigma);

    //! return a reasonable mask width of a gausian with sigma
    int getGaussWidth(double sigma);

private:
    int d_MaskWidth; //! the width of the gausian
    int d_SumOfMask; //! the sum of the mask element needed for normalization
    double *d_GaussArray; //! the gaussian array
    int *d_intGaussArray; //! the intiger gaussian array to accelerate
    double d_sigma; //! the standard deviation of the gaussian mask
};
}
#endif // __TUVISION_FILTERCGAUSS_H__
