/***************************************************************************
                          CCalibrator.h  -  description
                             -------------------
    begin                : 26.03.2001
    copyright            : (C) 2001 by Fabian Wenzel
    email                : tuvisionlib@tuvision.de, f.wenzel@gmx.net
    origin               : Diplomarbeit Fabian Wenzel

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library's internal version. It is   *
 *   not permitted to publish this file at this point of time as the       *
 *   algorithms are meant to be shared between TUHH Vision Systems         *
 *   members only.                                                         *
 *                                                                         *
 ***************************************************************************/
#ifndef __TUVISION_CCALIBRATOR_H__
#define __TUVISION_CCALIBRATOR_H__ 1

#include "TUVisionDef.h"
#include "newmat/newmat.h"
#include "tuv3d/TsaiCalibrationBackend.h"
#include "tuv3d/TsaiCameraModel.h"
#include <string>

namespace TUVision 
{

/*******************************************************************************\
 *                                                                              *
 * Calibration data consists of the (x,y,z) world coordinates of a feature      *
 * point	(in mm) and the corresponding coordinates (Xf,Yf) (in pixels) of the   *
 * feature point in the image.  Point count is the number of points in the data *
 * set.
 *
 *
 * Coplanar calibration:
 *
 * For coplanar calibration the z world coordinate of the data must be zero.
 * In addition, for numerical stability the coordinates of the feature points
 * should be placed at some arbitrary distance from the origin (0,0,0) of the
 * world coordinate system.  Finally, the plane containing the feature points
 * should not be parallel to the imaging plane.  A relative angle of 30 degrees
 * is recommended.
 *
 *
 * Noncoplanar calibration:
 *
 * For noncoplanar calibration the data must not lie in a single plane.
 *
\*******************************************************************************/
struct _CCTUVISION SCalibrationData {
    int       point_count;	/* [points] 	 */
    NEWMAT::RowVector *worldCoordinates; /* [mm]          */
    NEWMAT::RowVector *imageCoordinates; /* [pix]         */
};

struct _CCTUVISION SCalibrationError {
    double mean;
    double max;
    double stddev;
    double sse;
};

/**********************************************************/
//!  CCalibrator Class
/*! 
 *   This class is a wrapper for a modified version
 *   of Tsai's Camera Calibration algorithms by Reg Willson
 *
 *   \ingroup 3d
 *
 *   \author Fabian Wenzel
 */
/**********************************************************/

class _CCTUVISION CCalibrator {
public:
    CCalibrator();
    ~CCalibrator();
    bool Init(SCameraParameters *cameraParameters, SCalibrationConstants *calibrationConstants = 0); //!< Initialize Calibration parameters with camera parameters which can be found in technical specs
    bool Init(int width, int height); //!< Initialize calibration parameters with image extents
    bool Init(std::string cameraparameterfilename); //<! Initialize camera parameters from data file
    bool Save(std::string cameraparameterfilename);
    bool InitializeTsai(SCalibrationData *calibrationdata);
    bool Calibrate(SCalibrationData *calibrationdata);
    bool ExtrinsicCalibrate(SCalibrationData *calibrationdata);
    NEWMAT::Matrix *getNewRotationMatrix();
    NEWMAT::Matrix *getNewTranslationVector();
    NEWMAT::Matrix *getNewProjectionMatrix();
    void getTranslationVector(NEWMAT::ColumnVector &translationVector);
    void getProjectionMatrix(NEWMAT::Matrix &projectionMatrix);
    void getRotationMatrix(NEWMAT::Matrix &rotationMatrix);
    void getRTMatrix(NEWMAT::Matrix &Rt);
    void getAMatrix(NEWMAT::Matrix &A);
    void getCalibrationErrors(SCalibrationError *undistortedImagePlaneError,SCalibrationError *distortedImagePlaneError,SCalibrationError *objectSpaceError,SCalibrationError *normalizedCalibrationError);
    SCameraParameters *getCameraParameters() { return d_CameraParameters; }
    SCalibrationConstants *getCalibrationConstants() { return d_CalibrationConstants; }

private:
    bool                   d_initialized;
    bool                   d_calibrated;
    int                    d_NumOfCalibrationPoints;
    int                    d_NumOfRadialDistortionParameters;
    SCameraParameters     *d_CameraParameters;
    SCalibrationConstants *d_CalibrationConstants;
    SCalibrationData      *d_CalibrationData;
    TsaiCalibrator        *d_tsaiCalibrator;
};

} 

#endif // __TUVISION_CCALIBRATOR_H__
