/***************************************************************************
                          CIntrinsicMatrix.h  -  description
                             -------------------
    begin                : July 17 2002
    copyright            : (C) 2002 by Marco Grimm
    email                : grimm@tu-harburg.de

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#ifndef __TUVISION_CINTRINSICMATRIX_H
#define __TUVISION_CINTRINSICMATRIX_H

#include "newmat/newmat.h"
#include <vector>


namespace TUVision 
{
/***************************************************************************/
//! IntrinsicMatrix - Matrix with the intrinsic camera parameters
/*!
 *  This class is based on a NEWMAT matrix and includes the intrinsic 
 *  camera parameters of the pinhole camera model.
 *  (see Faugeras, Luong; 2001; The Geometry of Multiple Images, page 13 ) 
 *
 * \ingroup 3d
 *
 *  @author  Marco Grimm
 *
 ***************************************************************************/
class _CCTUVISION CIntrinsicMatrix : public NEWMAT::Matrix
{
public:
    //! Get the coordinate u0 from the principal point, matrix element(1,3)
    double getU0(){ return (*this)(1,3); };

    //! Set the coordinate u0 from the principal point, matrix element(1,3) 
    void   setU0( double u0 ){ (*this)(1,3) = u0; };

    //! Get the coordinate v0 from the principal point, matrix element(2,3) 
    double getV0(){     return (*this)(2,3); };

    //! Set the coordinate v0 from the principal point, matrix element(2,3) 
    void   setV0( double v0 ){ (*this)(2,3) = v0; };

    //! Get focal length alpha_u expressed in pixel units, matrix element(1,1)
    double getAlphaU(){         return (*this)(1,1); };

    //! Set focal length alpha_u expressed in pixel units, matrix element(1,1)
    void   setAlphaU( double alphaU ){ (*this)(1,1) = alphaU; };

    //! Get focal length alpha_v expressed in pixel units, matrix element(2,2)
    double getAlphaV(){         return (*this)(2,2); };

    //! Set focal length alpha_v expressed in pixel units, matrix element(2,2)
    void   setAlphaV( double alphaV ){ (*this)(2,2) = alphaV; };

    //! Get skew, parameter for non-orthogonal pixels, matrix element(1,2)
    double getSkew(){       return (*this)(1,2); };

    //! Set skew, parameter for non-orthogonal pixels, matrix element(1,2)
    void   setSkew( double skew ){ (*this)(1,2) = skew; };

    //! Standard Constructor initializes the matrix
    CIntrinsicMatrix() :
    Matrix(3,3)
    {
        (*this) << 1 << 0 << 0
            <<     0 << 1 << 0
            <<     0 << 0 << 1;
    };

    //! Constructor to copy an existing NEWMAT 3x3 matrix
    CIntrinsicMatrix(const NEWMAT::Matrix &P) : Matrix(3,3) {
        // if P has wrong size, SubMatrix throws an IncompatibleDimensionsException
        SubMatrix(1,3,1,3)=P;
    }

    //! Operator to assign an existing NEWMAT 3x3 matrix.
    CIntrinsicMatrix &operator=(const NEWMAT::Matrix &P)  {
        // if P has wrong size, SubMatrix throws an IncompatibleDimensionsException
        SubMatrix(1,3,1,3)=P;
        return (*this);
    }

    //! Destructor does nothing special
    virtual ~CIntrinsicMatrix(){};
};
}
#endif // ifndef CINTRINSICMATRIX_H
