/***************************************************************************
                          CGeneratorZoneplate.cpp  -  description
                             -------------------
    begin                : 19.11.2001
    copyright            : (C) 2001 by Thomas Beilicke, Fabian Wenzel
    email                : tuvisionlib@tuvision.de
    origin               : Digitale Bildverarbeitung 1, WS 2001/2001

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#include "tuv/CGeneratorZoneplate.h"
#include "tuv/CImageByteGray.h"
#include <cmath>

#ifndef M_PI
#define M_PI 3.1415926535897932384626433832795
#endif

#ifndef M_PI_4
#define M_PI_4 M_PI/4.0
#endif

namespace TUVision{

    CGeneratorZoneplate::CGeneratorZoneplate() : d_k0(0.0), d_kx(0.0),
        d_kx2(4.0), d_ky2(4.0), d_kxy(0.0),
        d_rLow(50), d_rHigh(200){
        }

        /*****************************************************************/ 
        /*! sinPlate constructs a sine-zone-plate with given dimension 
         *
         *  Mean amplitude 128, sine amplitude d_rHigh/2
         * 
         * \param width Width of the new image
         * \param height Height of the new image
         *
         * \return A new zone plate image
         *
         * \warning The returned image has to be deleted later by hand
         *****************************************************************/ 
        CImageByteGray* CGeneratorZoneplate::sinPlate(int width, int height) {

            CImageByteGray *image=new CImageByteGray(width, height);

            double amplitude = double (d_rHigh)/2.0;
            double x, y, phase;
            for (int yi=0; yi<height;yi++) {
                unsigned char* row=image->getRow(yi);
                y=yi/(height-1.0); 
                for (int xi=0; xi<width; xi++) {
                    x=xi/(width-1.0);  // the For-loop counts 0..(width-1) !!
                    phase=2.0*M_PI*(d_k0 + d_kx*x + d_ky*y + 0.5*d_kx2*x*x + 0.5*d_ky2*y*y + d_kxy*x*y);
                    row[xi]=(int)(128.0 + amplitude*sin(phase));
                }
            }
            return image;
        }

        /*****************************************************************/ 
        /*! rampPlate constructs a Ramp-Plate with given dimension
         * 
         * \param width Width of the new image
         * \param height Height of the new image
         *
         * \return A new zone plate image
         * \warning The returned image has to be deleted later by hand
         *
         *****************************************************************/ 
        CImageByteGray* CGeneratorZoneplate::rampPlate(int width, int height) {
            CImageByteGray *image=new CImageByteGray(width, height);

            double x, y, phase;
            for (int yi=0; yi<height;yi++) {
                unsigned char* row=image->getRow(yi);
                y=yi/(height-1.0);
                for (int xi=0; xi<width; xi++) {
                    x=xi/(width-1.0);    // the For-loop counts 0..(width-1) !!
                    phase=2*M_PI*(d_k0 + d_kx*x + d_ky*y + 0.5*d_kx2*x*x + 0.5*d_ky2*y*y + d_kxy*x*y);
                    row[xi]=(int)(ramp(phase)); 
                }
            }
            return image;
        }

        /*****************************************************************/ 
        /*! setAmplitudeInterval sets the amplitude interval
         *  for the image function
         * 
         * \param low Lower bound for the interval
         * \param high Higher bound for the interval
         *
         *****************************************************************/ 
        void CGeneratorZoneplate::setAmplitudeInterval(unsigned char low, unsigned char high) {
            d_rLow =low;
            d_rHigh=high;
        }

        /* ramp calculates the ramp-function */
        double CGeneratorZoneplate::ramp(double phase) {  
            /* Make the phaseument 2M_PI-periodic */
            while (phase>=2.0*M_PI)
                phase-=2.0*M_PI;
            while (phase<0)
                phase+=2.0*M_PI;

            if (phase>=5.0*M_PI_4)    // 5M_PI_4 < phase < 2M_PI
                return d_rLow;
            if (phase>=M_PI)                    // M_PI < phase < 5M_PI_4
                return d_rHigh-(d_rHigh - d_rLow)* (phase-M_PI)*4.0/M_PI;
            if (phase>=M_PI_4)            // M_PI_4 < phase < M_PI
                return d_rHigh;
            else                                    // 0 < phase < M_PI
                return d_rLow+phase*(d_rHigh - d_rLow) * 4.0/M_PI;
        }
}
