/***************************************************************************
                          CFileWriteTiff.cpp  -  description
                             -------------------
    begin                : 01.01.2001
    copyright            : (C) 2001 by Marco Budde
    email                : tuvisionlib@tuvision.de
    origin               : Diplomarbeit Marco Budde

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#include "tuv/CFileWriteTiff.h"

#include <string>

#include <tiffio.h>
#include "tuv/CImageByteBilevel.h"
#include "tuv/CImageByteGray.h"
#include "tuv/CImageByteRGB.h"
#include "tuv/CException.h"
#include "tuv/CExceptionFile.h"
#include "tuv/CImageAttributes.h"


using namespace std;
using namespace TUVision;



CFileWriteTiff::CFileWriteTiff (void)
{
    d_filename = "";
    d_tif_hd = NULL;
    d_compression = false;
    TIFFSetErrorHandler (NULL);
}


CFileWriteTiff::~CFileWriteTiff()
{
    if (d_tif_hd != NULL)
        TIFFClose (d_tif_hd);
}


void CFileWriteTiff::open (const string &filename)
throw (CException)
{
    d_filename = filename;

    d_tif_hd = TIFFOpen (d_filename.c_str(), "w");
    if (d_tif_hd == NULL)
        throw CExceptionFileOpen (d_filename);
}


void CFileWriteTiff::close (void)
throw()
{
    if (d_tif_hd != NULL)
    {
        TIFFClose (d_tif_hd);
        d_tif_hd = NULL;
    }
}


/******************************************************/
/*! 
 *   Saves the following CImageAttributes in the
 *   TIFF file opened by this class: xdpi, ydpi,
 *   description.
 *
 *   \param img image to be stored
 *   \return success (true), error (false)
 */
/******************************************************/

bool CFileWriteTiff::writeAttributes (const CImage &img)
{
    float  xdpi, ydpi;
    string description;  


    if (TIFFSetField (d_tif_hd, TIFFTAG_RESOLUTIONUNIT, 
        RESUNIT_INCH) != 1)
        return (false);

    xdpi = (float) img.getAttributes().getDoubleAttribute ("xdpi");
    if (xdpi > 0.0)
        if (TIFFSetField (d_tif_hd, TIFFTAG_XRESOLUTION, xdpi) != 1)
            return (false);

    ydpi = (float) img.getAttributes().getDoubleAttribute ("ydpi");
    if (ydpi > 0.0)
        if (TIFFSetField (d_tif_hd, TIFFTAG_YRESOLUTION, ydpi) != 1)
            return (false);

    description = 
        img.getAttributes().getStringAttribute ("description");
    if (description != "")  
        if (TIFFSetField (d_tif_hd, TIFFTAG_IMAGEDESCRIPTION, 
            description.c_str()) != 1)
            return (false);

    return (true);
}


/******************************************************/
/*! 
 *   This function writes the given image (8 bit gray)
 *   to TIFF graphic file. To store more than one
 *   image in the graphic file, simply call this 
 *   function several times.
 *
 *   \param img image to be saved
 */
/******************************************************/

void CFileWriteTiff::write (const CImageByteGray &img)
throw (CException)
{
    unsigned char  *img_buffer;
    int            i;
    unsigned int   size_x, size_y;


    // TIFF file handle valid?
    if (d_tif_hd == NULL)
        throw CExceptionFileWrite (d_filename);

    size_x = img.getW();
    size_y = img.getH();

    // Write header.
    if (TIFFSetField (d_tif_hd, TIFFTAG_IMAGEWIDTH, size_x) != 1)
        throw CExceptionFileWrite (d_filename);
    if (TIFFSetField (d_tif_hd, TIFFTAG_IMAGELENGTH, size_y) != 1)
        throw CExceptionFileWrite (d_filename);
    if (TIFFSetField (d_tif_hd, TIFFTAG_PLANARCONFIG, 
        PLANARCONFIG_CONTIG) != 1)
        throw CExceptionFileWrite (d_filename);
    if (TIFFSetField (d_tif_hd, TIFFTAG_BITSPERSAMPLE, 8) != 1)
        throw CExceptionFileWrite (d_filename);
    if (TIFFSetField (d_tif_hd, TIFFTAG_SAMPLESPERPIXEL, 1) != 1)
        throw CExceptionFileWrite (d_filename);
    if (TIFFSetField (d_tif_hd, TIFFTAG_PHOTOMETRIC, 
        PHOTOMETRIC_MINISBLACK) != 1)
        throw CExceptionFileWrite (d_filename);
    if (TIFFSetField (d_tif_hd, TIFFTAG_ORIENTATION, 
        ORIENTATION_TOPLEFT) != 1)
        throw CExceptionFileWrite (d_filename);

    // Enable compression?
    if (d_compression)
        if (TIFFSetField (d_tif_hd, TIFFTAG_COMPRESSION,  
            COMPRESSION_LZW) != 1)
            throw CExceptionFileWrite (d_filename);

    // Save CImage attributes.
    if (writeAttributes (img) == false)
        throw CExceptionFileWrite (d_filename);

    // Save image itself.
    for (i = 0; i < img.getH(); i++)
    {
        img_buffer = const_cast <unsigned char*> (img.getRow (i));
        if (TIFFWriteScanline (d_tif_hd, img_buffer, i) == -1)
            throw CExceptionFileWrite (d_filename);
    }

    // Create new TIFF directory.
    if (TIFFWriteDirectory (d_tif_hd) != 1)
        throw CExceptionFileWrite (d_filename);
}



/******************************************************/
/*! 
 *   This function writes the given image (bilevel)
 *   to TIFF graphic file. To store more than one
 *   image in the graphic file, simply call this 
 *   function several times.
 *
 *   \param img image to be saved
 */
/******************************************************/

void CFileWriteTiff::write (const CImageByteBilevel &img)
throw (CException, exception)
{
    const unsigned char  *img_buffer;
    unsigned char        *my_img_buffer;
    int                  i, x;
    unsigned int         size_x, size_y;


    // TIFF file handle valid?
    if (d_tif_hd == NULL)
        throw CExceptionFileWrite (d_filename);

    size_x = img.getW();
    size_y = img.getH();

    // Write header.
    if (TIFFSetField (d_tif_hd, TIFFTAG_IMAGEWIDTH, size_x) != 1)
        throw CExceptionFileWrite (d_filename);
    if (TIFFSetField (d_tif_hd, TIFFTAG_IMAGELENGTH, size_y) != 1)
        throw CExceptionFileWrite (d_filename);
    if (TIFFSetField (d_tif_hd, TIFFTAG_PLANARCONFIG, 
        PLANARCONFIG_CONTIG) != 1)
        throw CExceptionFileWrite (d_filename);
    if (TIFFSetField (d_tif_hd, TIFFTAG_BITSPERSAMPLE, 1) != 1)
        throw CExceptionFileWrite (d_filename);
    if (TIFFSetField (d_tif_hd, TIFFTAG_SAMPLESPERPIXEL, 1) != 1)
        throw CExceptionFileWrite (d_filename);
    if (TIFFSetField (d_tif_hd, TIFFTAG_PHOTOMETRIC, 
        PHOTOMETRIC_MINISBLACK) != 1)
        throw CExceptionFileWrite (d_filename);
    if (TIFFSetField (d_tif_hd, TIFFTAG_ORIENTATION, 
        ORIENTATION_TOPLEFT) != 1)
        throw CExceptionFileWrite (d_filename);

    // Enable compression?
    if (d_compression)
        if (TIFFSetField (d_tif_hd, TIFFTAG_COMPRESSION,  
            COMPRESSION_CCITTRLE) != 1)
            throw CExceptionFileWrite (d_filename);

    // Save CImage attributes.
    if (writeAttributes (img) == false)
        throw CExceptionFileWrite (d_filename);

    // Create local image buffer.
    my_img_buffer = new unsigned char [img.getW()/8 + 1];
    my_img_buffer[img.getW()/8] = 0;

    // Save image itself.
    for (i = 0; i < img.getH(); i++)
    {
        img_buffer = img.getRow (i);

        // Convert the image to the right format.
        for (x = 0; x < img.getW(); x++)
        {
            if (img_buffer[x] == 0)
                my_img_buffer[x/8] &= ~(1 << (7-(x % 8)));
            else
                my_img_buffer[x/8] |= (1 << (7-(x % 8)));
        }

        if (TIFFWriteScanline (d_tif_hd, my_img_buffer, i) == -1)
            throw CExceptionFileWrite (d_filename);
    }

    delete [] my_img_buffer;

    // Create new TIFF directory.
    if (TIFFWriteDirectory (d_tif_hd) != 1)
        throw CExceptionFileWrite (d_filename);
}


/******************************************************/
/*! 
 *   This function writes the given image (8 bit RGB)
 *   to TIFF graphic file. To store more than one
 *   image in the graphic file, simply call this 
 *   function several times.
 *
 *   \param img image to be saved
 */
/******************************************************/

void CFileWriteTiff::write (const CImageByteRGB &img)
throw (CException)
{
    unsigned char  *img_buffer;
    int            i;
    unsigned int   size_x, size_y;

    // TIFF file handle valid?
    if (d_tif_hd == NULL)
        throw CExceptionFileWrite (d_filename);

    size_x = img.getW();
    size_y = img.getH();

    // Write header.
    if (TIFFSetField (d_tif_hd, TIFFTAG_IMAGEWIDTH, size_x) != 1)
        throw CExceptionFileWrite (d_filename);
    if (TIFFSetField (d_tif_hd, TIFFTAG_IMAGELENGTH, size_y) != 1)
        throw CExceptionFileWrite (d_filename);
    if (TIFFSetField (d_tif_hd, TIFFTAG_PLANARCONFIG, 
        PLANARCONFIG_CONTIG) != 1)
        throw CExceptionFileWrite (d_filename);
    if (TIFFSetField (d_tif_hd, TIFFTAG_BITSPERSAMPLE, 8) != 1)
        throw CExceptionFileWrite (d_filename);
    if (TIFFSetField (d_tif_hd, TIFFTAG_SAMPLESPERPIXEL, 3) != 1)
        throw CExceptionFileWrite (d_filename);
    if (TIFFSetField (d_tif_hd, TIFFTAG_PHOTOMETRIC, 
        PHOTOMETRIC_RGB) != 1)
        throw CExceptionFileWrite (d_filename);
    if (TIFFSetField (d_tif_hd, TIFFTAG_ORIENTATION, 
        ORIENTATION_TOPLEFT) != 1)
        throw CExceptionFileWrite (d_filename);

    // Enable compression?
    if (d_compression)
        if (TIFFSetField (d_tif_hd, TIFFTAG_COMPRESSION,  
            COMPRESSION_LZW) != 1)
            throw CExceptionFileWrite (d_filename);

    // Save CImage attributes.
    if (writeAttributes (img) == false)
        throw CExceptionFileWrite (d_filename);

    // Save image itself.
    for (i = 0; i < img.getH(); i++)
    {
        img_buffer = const_cast <unsigned char*> (img.getRow (i));
        if (TIFFWriteScanline (d_tif_hd, img_buffer, i) == -1)
            throw CExceptionFileWrite (d_filename);
    }

    // Create new TIFF directory.
    if (TIFFWriteDirectory (d_tif_hd) != 1)
        throw CExceptionFileWrite (d_filename);
}


/******************************************************/
/*!
 *   Save the following images with compression
 *   turned on or off.
 *
 *   \param Compression compression on/off?
 *
 *   \warning The latest versions of the libtiff don't
 *            support the LZW compression due to patent 
 *            problems.
 */
/******************************************************/

void CFileWriteTiff::setParameters (bool compression)
throw()
{
    d_compression = compression;
}
