/***************************************************************************
                          CCurveSearch.cpp  -  description
                             -------------------
    begin                : 1.1.2000
    copyright            : (C) 2000 by Islam Shdaifat
    email                : tuvisionlib@tuvision.de
    origin               : PhD - Thesis Islam Shdaifat
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#include "tuv2d/CCurveSearch.h"
#include "tuv/CImageByteGray.h"
#include "tuv2d/CShape.h"
#include <iostream>

using namespace TUVision;
using namespace std;

CCurveSearch::~CCurveSearch()
{
}


//! construct a CurveSearch object
CCurveSearch::CCurveSearch(vector< CCurve > vecSrCrs,
                           double dThr0 ,
                           double dThr1 )
{
    vecCurvesLUT = vecSrCrs;
    itCurveStart = vecCurvesLUT.begin();
    itCurveEnd = vecCurvesLUT.end();
    d_curve = &(*itCurveStart);
    setThreshold(dThr0, dThr1);
    delCurveFromImge = false;
    fillInLinePoints();
    dp = 0;
    ImageSize = 0;
};

CCurveSearch::CCurveSearch(CShape& shape,
                           double dThr0 ,
                           double dThr1 )
{

    vecCurvesLUT.insert(vecCurvesLUT.begin(),
        shape.vecCurvesLUT.begin(),shape.vecCurvesLUT.end());
    delCurveFromImge = false;
    fillInLinePoints();
    dp = 0;
    ImageSize = 0;
    //setThreshold(dThr0, dThr1);
}

void
CCurveSearch::insertShape(CShape& shape)
{
    vecCurvesLUT.insert(vecCurvesLUT.end(),shape.vecCurvesLUT.begin() ,shape.vecCurvesLUT.end());
    getRidOfDuplicatedCurves();
}

/**********************************************************/
/*! 
 * set the search curve pointer
 *
 * \param crv a pointer to the curve to be search in the image
 */
/**********************************************************/
void
CCurveSearch::setSearchCurveVector( CCurve *crv)
{
    d_curve = crv;
};


/**********************************************************/
/*! 
 * set the thresholds of search
 *
 * \param d_Thr0 the max non hit points
 * \param d_Thr1 the min hit points
 */
/**********************************************************/
void
CCurveSearch::setThreshold(double dThr0 , double dThr1 )
{
    d_Max_NonHitCount = d_curve->vecCurvePoints.size();
    d_Min_HitCount    = d_curve->vecCurvePoints.size();
};



/**********************************************************/
/*! 
 * search the Multi Curves in the Image
 *
 * \param img a pointer to the image to search in
 */
/**********************************************************/
void
CCurveSearch::search(CImageByteGray *img, int x0, int y0)
{ 
    setSearchInImage( *img ); // set the search in the image
    vecSearchOut.clear(); // clear the output vector

    int w = d_image->getW() - x0; // image width
    int h = d_image->getH() - y0; // image height

    // use setWindow() to be inside the image or Pad the image
    // if necessary, this shoul be done before SEARCH
    for(d_y = 0 ; d_y < h ; d_y++) // loop through height
    { 
        d_image_ptr = d_image->getRow(d_y) + x0;
        for(d_x = x0 ; d_x < w  ; d_x++) // loop through width
        {
            searchSpecial(); // should be implemented for each shape seperately
            d_image_ptr++;
        } // end x, width
    } // end y, height
}


/**********************************************************/
/*! 
 * serch the curves starting at itCurveStart to itCurveEnd
 * in the image
 */
/**********************************************************/
void
CCurveSearch::checkCurves_itStart_To_itEnd() {

    // go through the specified curves start with
    // itCurveStart and end at itCurveEnd
    for( d_it_crv = itCurveStart; d_it_crv < itCurveEnd; ++d_it_crv) {
        d_curve = &(*d_it_crv); // set the internal search curve
        // you may need to set the d_max_0 and d_min_1 here if different
        // curve thresholds are needed.
        //setThreshold();

        getHitAndNonHitCount(); // count the hit and non hit points
        if( isSuffecientHitPoints() )
            writeToFeatureStream();
    }

}


/**********************************************************/
/*! 
 * count hit and non hit points
 */
/**********************************************************/
void
CCurveSearch::getHitAndNonHitCount() {

    int i;
    unsigned char tmp;
    int length = d_curve->vecCurvePoints.size();

    d_Hit = 0; // reset counters
    d_NonHit = 0;

    for( i = 0 ; i < length ; i++) {
        // the value of the image at that point
        tmp = d_image->getPixel(d_curve->vecCurvePoints[i].x + d_x , 
            d_curve->vecCurvePoints[i].y + d_y);

        if( tmp == 255) { // it is a hit point
            d_Hit++;
        } // hit point
        else { // non hit point
            d_NonHit++;
            if(d_NonHit > d_Max_NonHitCount ) {
                // it is better to break now rather to continue searching
                break;
            }
        }// non hit point
    } // end the curve
}

/**********************************************************/
/*! 
 * check if the Hit Points > threshold
 */
/**********************************************************/
bool
CCurveSearch::isSuffecientHitPoints() {

    if(d_NonHit <= d_Max_NonHitCount && d_Hit >= d_Min_HitCount)
        return true;
    return false;
}

/**********************************************************/
/*!
 * write the output to the OutputVector 
 */
/**********************************************************/
void
CCurveSearch::writeToFeatureStream() {

    //write to output
    COut tmpout;
    if(d_Hit >= d_Min_HitCount ) {
        /*
        cout
        <<"\n is found at x:          "<< d_x
        <<"\n and y:                  "<< d_y
        <<"\n with hit points:        "<< d_Hit
        <<"\n with NON hit points:    "<< d_NonHit
        <<"\n the curve pointer:      "<< d_it_crv
        <<"\n"
        ;
        */
        tmpout.crv = &(*d_it_crv);
        tmpout.x = d_x;
        tmpout.y = d_y;
        tmpout.xf = d_it_crv->CurveEndPoint.x;
        tmpout.yf = d_it_crv->CurveEndPoint.y;
        tmpout.Hits = d_Hit;
        tmpout.NonHits = d_NonHit;

        vecSearchOut.push_back(tmpout); // insert it in the output curve
    } // a good curve was found
}

/**********************************************************/
/*!
 * draw a curve in the image at the givn location
 * 
 * \warning
 * This function should be used carefully, whether you draw
 * inside the same image which you search in.
 *
 * \param img the image to draw the curve in
 * \param crv the curve to be drawn in the image
 * \param x the curve to be drawn in the image at x coord.
 * \param y the curve to be drawn in the image at y coord.
 * \param color the gray level to be set in the image != 255
 */
/**********************************************************/
void
CCurveSearch::drawCurveInImage(CImageByteGray& img, CCurve& crv,
                               int x, int y, int color)
{
    unsigned char tmp = color;

    int im_w = img.getW();
    int im_h = img.getH();
    for(int i = 0 ; i < crv.CurveLength ; i++)
    {
        int k = crv.vecCurvePoints[i].y + y;
        int j = crv.vecCurvePoints[i].x + x;
        if( k > 0 && k < im_h && j > 0 && j < im_w)
            // the value of the image at that point
            img.setPixel(j , 
            k, tmp);
    } // end the curve

}


void
CCurveSearch::setSearchInImage(CImageByteGray &image)
{
    d_image = &image;
    h_im = image.getH();
    w_im = image.getW();
    int delta = abs( image.getRow(1) - image.getRow(0) ); // the difference between two rows in image
    if(dp != delta)
    {
        dp = delta;
        fillInLinePoints();
        /*
        for(i = 0; i < size_Mask; i++)
        x_y_Mask[i] = x_Mask[i]  + y_Mask[i]*dp;
        */
    }
    if(ImageSize != (h_im * w_im))
    {
        /*
        ImageSize = image.getW() * image.getH();
        delete [] r;
        delete [] cgx;
        delete [] cgy;

        r   = new int[ImageSize];
        cgx = new int[ImageSize];
        cgy = new int[ImageSize];
        */
    }

};

