/***************************************************************************
                          CFilterConvolution.cpp  -  description
                             -------------------
    begin                : 01.01.2001
    copyright            : (C) 2001 by Marco Budde
    email                : tuvisionlib@tuvision.de
    origin               : Diplomarbeit Marco Budde

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#include "tuv2d/CFilterConvolution.h"
#include "tuv/CImageByteGray.h"
#include "tuv/CImageIntGray.h"
#include "tuv/CImageDoubleGray.h"


//using namespace std;
using namespace TUVision;


CFilterConvolution::CFilterConvolution()
{
    d_matrix_c = 0;
    d_matrix_f = 0;
}


CFilterConvolution::~CFilterConvolution()
{
    delete [] d_matrix_c;
    delete [] d_matrix_f;
}


/**********************************************************/
/*! 
 *   Sets the convolution matrix to be used by this
 *   filter object. Each filter object stores its own
 *   copy of the convolution matrix. It is possible to
 *   set a scaling factor. The resulting image of the
 *   convolution is divided by this scaling factor:
 *
 *   \f[ \frac{1}{scaling} 
 *       \left[\begin{array}{ccc} 
 *             matrix_{00} & \cdots & matrix_{0x} \\
 *             \vdots      &        & \vdots  \\
 *             matrix_{y0} & \cdots & matrix_{xy}
 *       \end{array}\right]  \f]
 *
 *   \param x         width of the matrix
 *   \param y         height of the matrix
 *   \param matrix_c  the convolution matrix
 *   \param scaling_c scaling factor
 */
/**********************************************************/

void 
CFilterConvolution::setParameters (int x, int y, 
                                   const char *matrix_c, 
                                   char scaling_c)
{
    int i;


    delete [] d_matrix_c;
    delete [] d_matrix_f;

    d_matrix_c = new char [x*y];
    d_matrix_f = new float [x*y];

    d_width = x;
    d_height = y;
    d_matrix_float = false;

    for (i=0; i < x * y; i++)
    {
        d_matrix_c[i] = matrix_c[i];
        d_matrix_f[i] = (float) matrix_c[i];
    }

    d_scaling_c = scaling_c;
    d_scaling_f = (float) scaling_c;
}


/**********************************************************/
/*! 
 *   Sets the convolution matrix to be used by this
 *   filter object. Each filter object stores its own
 *   copy of the convolution matrix. It is possible to
 *   set a scaling factor. The resulting image of the
 *   convolution is divided by this scaling factor:
 *
 *   \f[ \frac{1}{scaling} 
 *       \left[\begin{array}{ccc} 
 *             matrix_{00} & \cdots & matrix_{0x} \\
 *             \vdots      &        & \vdots  \\
 *             matrix_{y0} & \cdots & matrix_{xy}
 *       \end{array}\right]  \f]
 *
 *   \param x         width of the matrix
 *   \param y         height of the matrix
 *   \param matrix_f  the convolution matrix
 *   \param scaling_f scaling factor
 */
/**********************************************************/

void 
CFilterConvolution::setParameters (int x, int y, 
                                   const float *matrix_f,
                                   float scaling_f)
{
    int i;


    delete [] d_matrix_c;
    delete [] d_matrix_f;

    d_matrix_c = new char [x*y];
    d_matrix_f = new float [x*y];

    d_width = x;
    d_height = y;
    d_matrix_float = true;

    for (i=0; i < x * y; i++)
    {
        d_matrix_c[i] = (char) matrix_f[i];
        d_matrix_f[i] = matrix_f[i];
    }

    d_scaling_c = (char) scaling_f;
    d_scaling_f = scaling_f;
}


/**********************************************************/
/*! 
 *   Filter the given image (byte gray) using the 
 *   specified convolution matrix. Results larger/smaller
 *   than 255/0 are clipped.
 *
 *   \param img the image to be filtered
 */
/**********************************************************/

void CFilterConvolution::filter (CImageByteGray &img)
{
    unsigned char  **row;
    unsigned char  *res_row;
    int            x, y, i, x_m, y_m;
    int            result;
    CImageByteGray *orig_img;


    // Tests whether the user has specified a convolution
    // matrix.
    if ((d_matrix_c == 0) || (d_matrix_f == 0))
        return;

    // Allocate row pointer array.
    row = new unsigned char * [d_height]; 

    // Create a copy of the image.
    orig_img = new CImageByteGray (img);

    // Clear old image (necessary because of the boundary).
    img.clearImage();

    for (y = 0; y < orig_img->getH() - d_height + 1; y++)
    {
        // Get row pointers of input image needed to calculate
        // one row in the output image.
        for (i = 0; i < d_height; i++)
            row[i] = orig_img->getRow (y + i);

        // Get row pointer of the output image.
        res_row = img.getRow (y + d_height/2);

        for (x = 0; x < orig_img->getW() - d_width + 1; x++)
        {
            // Calculate one pixel of the output image.
            result = 0;
            for (y_m = 0; y_m < d_height; y_m++)
                for (x_m = 0; x_m < d_width; x_m++)
                {
                    if (d_matrix_float == false)
                        result += row[y_m][x + x_m] * 
                        d_matrix_c[y_m * d_width + x_m];
                    else
                        result += (int) (row[y_m][x + x_m] * 
                        d_matrix_f[y_m * d_width + x_m]);
                }

                // scaling
                if (d_matrix_float == false)
                    result /= d_scaling_c;
                else
                    result = int (result / d_scaling_f);

                // clipping
                if (result > 255)
                    res_row [x + d_width/2] = 255;
                else if (result < 0)
                    res_row [x + d_width/2] = 0;
                else
                    res_row [x + d_width/2] = result;
        }
    } 

    delete [] row;
    delete orig_img;
}  


/**********************************************************/
/*! 
 *   Filter the given image (int gray) using the 
 *   specified convolution matrix. 
 *
 *   \param img the image to be filtered
 */
/**********************************************************/

void CFilterConvolution::filter (CImageIntGray &img)
{
    int            **row;
    int            *res_row;
    int            x, y, i, x_m, y_m;
    int            result;
    CImageIntGray  *orig_img;


    // Tests whether the user has specified a convolution
    // matrix.
    if ((d_matrix_c == 0) || (d_matrix_f == 0))
        return;

    // Allocate row pointer array.
    row = new int * [d_height]; 

    // Create a copy of the image.
    orig_img = new CImageIntGray (img);

    // Clear old image (necessary because of the boundary).
    img.clearImage();

    for (y = 0; y < orig_img->getH() - d_height + 1; y++)
    {
        // Get row pointers of input image needed to calculate
        // one row in the output image.
        for (i = 0; i < d_height; i++)
            row[i] = orig_img->getRow (y + i);

        // Get row pointer of the output image.
        res_row = img.getRow (y + d_height/2);

        for (x = 0; x < orig_img->getW() - d_width + 1; x++)
        {
            // Calculate one pixel of the output image.
            result = 0;
            for (y_m = 0; y_m < d_height; y_m++)
                for (x_m = 0; x_m < d_width; x_m++)
                {
                    if (d_matrix_float == false)
                        result += row[y_m][x + x_m] * 
                        d_matrix_c[y_m * d_width + x_m];
                    else
                        result += (int) (row[y_m][x + x_m] * 
                        d_matrix_f[y_m * d_width + x_m]);
                }

                // scaling
                if (d_matrix_float == false)
                    result /= d_scaling_c;
                else
                    result = int (result / d_scaling_f);

                res_row [x + d_width/2] = result;
        }
    } 

    delete [] row;
    delete orig_img;
}


/**********************************************************/
/*! 
 *   Filter the given image (double gray) using the 
 *   specified convolution matrix. 
 *
 *   \param img the image to be filtered
 */
/**********************************************************/

void CFilterConvolution::filter (CImageDoubleGray &img)
{
    double            **row;
    double            *res_row;
    int               x, y, i, x_m, y_m;
    double            result;
    CImageDoubleGray  *orig_img;


    // Tests whether the user has specified a convolution
    // matrix.
    if ((d_matrix_c == 0) || (d_matrix_f == 0))
        return;

    // Allocate row pointer array.
    row = new double * [d_height]; 

    // Create a copy of the image.
    orig_img = new CImageDoubleGray (img);

    // Clear old image (necessary because of the boundary).
    img.clearImage();

    for (y = 0; y < orig_img->getH() - d_height + 1; y++)
    {
        // Get row pointers of input image needed to calculate
        // one row in the output image.
        for (i = 0; i < d_height; i++)
            row[i] = orig_img->getRow (y + i);

        // Get row pointer of the output image.
        res_row = img.getRow (y + d_height/2);

        for (x = 0; x < orig_img->getW() - d_width + 1; x++)
        {
            // Calculate one pixel of the output image.
            result = 0;
            for (y_m = 0; y_m < d_height; y_m++)
                for (x_m = 0; x_m < d_width; x_m++)
                {
                    result += row[y_m][x + x_m] * 
                        d_matrix_f[y_m * d_width + x_m];
                }

                // Save pixel and scaling
                res_row [x + d_width/2] = result / d_scaling_f;
        }
    } 

    delete [] row;
    delete orig_img;
}
