/***************************************************************************
                          CFilterHistogram.cpp  -  description
                             -------------------
    begin                : 01.01.2001
    copyright            : (C) 2001 by Marco Budde
    email                : tuvisionlib@tuvision.de
    origin               : Diplomarbeit Marco Budde

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#include "tuv2d/CFilterHistogram.h"
#include "tuv/CImageByteGray.h"
#include "tuv/CImageIntGray.h"
#include <cmath>
#include <climits>


//using namespace std;
using namespace TUVision;


const static int MY_INT_MAX = 65535;



/**********************************************************/
/*!
 *   Constructor.
 */
/**********************************************************/

CFilterHistogram::CFilterHistogram()
{
    d_streching = false;
    d_gamma_correction = false;
    d_compression = false;
}


/**********************************************************/
/*!
 *   Enables streching (step 1) and sets the black and
 *   white points in the image to be filtered. Pixels
 *   with values smaller/larger than the black/white
 *   points are clipped to the black/white point values.
 *
 *   \param black  black point
 *   \param white  white point
 */
/**********************************************************/

void
CFilterHistogram::setStreching (int black, int white)
{
    d_strech_black = black;
    d_strech_white = white;
    d_streching = true;
}


/**********************************************************/
/*!
 *  Enables gamma correction (step 2) and sets the
 *  gamma correction value.
 *
 *  \param gamma  gamma value
 */
/**********************************************************/

void
CFilterHistogram::setGammaCorrection (double gamma)
{
    d_gamma = gamma;
    d_gamma_correction = true;
}


/**********************************************************/
/*!
 *  Enables compression (step 3) and sets the min and
 *  max values in the filtered output image.
 *
 *  \param black  min value
 *  \param white  max value
 */
/**********************************************************/

void
CFilterHistogram::setCompression (int black, int white)
{
    d_comp_black = black;
    d_comp_white = white;
    d_compression = true;
}


/**********************************************************/
/*!
 *  Disables all three steps: streching, gamma correction,
 *  and compression.
 */
/**********************************************************/

void
CFilterHistogram::reset (void)
{
    d_streching = false;
    d_gamma_correction = false;
    d_compression = false;
}


void
CFilterHistogram::filter (CImageByteGray &img)
{
    unsigned char *row;
    int           x, y;
    unsigned char lut[256];
    double        tmp;


    // Calculate LUT for streching the histogram.
    if (d_streching == true)
    {
        for (x = 0; x < 256; x++)
        {

            tmp = double ((x -  d_strech_black) * UCHAR_MAX) /
                double (d_strech_white - d_strech_black);

            if (tmp > 255.0)
                lut[x] = 255; 
            else if (tmp < 0.0)
                lut[x] = 0;
            else
                lut[x] = (unsigned char) tmp;
        }
    }
    else
    {
        for (x = 0; x < 256; x++)
            lut[x] = x;
    }

    // Calculate LUT for gamma correction.
    if (d_gamma_correction == true)
    {
        for (x = 0; x < 256; x++)
        {
            tmp = pow ((double)lut[x], 1.0 / d_gamma) / 
                pow (255.0, 1.0 / d_gamma) * 255.0;

            if (tmp > 255.0)
                lut[x] = 255; 
            else if (tmp < 0.0)
                lut[x] = 0;
            else
                lut[x] = (unsigned char) tmp;
        }
    }

    // Calculate LUT for compression of the histogram.
    if (d_compression == true)
    {
        for (x = 0; x < 256; x++)
        {

            tmp = double (lut[x] * (d_comp_white - d_comp_black)) /
                double (UCHAR_MAX) + double (d_comp_black);

            if (tmp > 255.0)
                lut[x] = 255; 
            else if (tmp < 0.0)
                lut[x] = 0;
            else
                lut[x] = (unsigned char) tmp;
        }
    }


    // Filter the image using the LUT. 
    for (y = 0; y < img.getH(); y++)
    {
        row = img.getRow(y);

        for (x = 0; x < img.getW(); x++)
            row[x] = lut[row[x]];
    }  
}


void
CFilterHistogram::filter (CImageIntGray &img)
{
    int           *row;
    int           x, y;
    int           lut[MY_INT_MAX + 1];
    double        tmp;


    // Calculate LUT for streching the histogram.
    if (d_streching == true)
    {
        for (x = 0; x < MY_INT_MAX + 1 ; x++)
        {

            tmp = double ((x -  d_strech_black) * MY_INT_MAX) /
                double (d_strech_white - d_strech_black);

            if (tmp > double (MY_INT_MAX))
                lut[x] = MY_INT_MAX; 
            else if (tmp < 0.0)
                lut[x] = 0;
            else
                lut[x] = (int) tmp;
        }
    }
    else
    {
        for (x = 0; x < MY_INT_MAX + 1; x++)
            lut[x] = x;
    }

    // Calculate LUT for gamma correction.
    if (d_gamma_correction == true)
    {
        for (x = 0; x < MY_INT_MAX + 1; x++)
        {
            tmp = pow ((double)lut[x], 1.0 / d_gamma) / 
                pow ((double)MY_INT_MAX, 1.0 / d_gamma) * MY_INT_MAX;

            if (tmp > double (MY_INT_MAX))
                lut[x] = MY_INT_MAX; 
            else if (tmp < 0.0)
                lut[x] = 0;
            else
                lut[x] = (int) tmp;
        }
    }

    // Calculate LUT for compression of the histogram.
    if (d_compression == true)
    {
        for (x = 0; x < MY_INT_MAX + 1; x++)
        {

            tmp = double (lut[x] * (d_comp_white - d_comp_black)) /
                double (MY_INT_MAX) + double (d_comp_black);

            if (tmp > double (MY_INT_MAX))
                lut[x] = MY_INT_MAX; 
            else if (tmp < 0.0)
                lut[x] = 0;
            else
                lut[x] = (int) tmp;
        }
    }


    // Filter the image using the LUT. 
    for (y = 0; y < img.getH(); y++)
    {
        row = img.getRow(y);

        for (x = 0; x < img.getW(); x++)
        {
            if (row[x] < 0)
                row[x] = 0;
            else if (row[x] > MY_INT_MAX)
                row[x] = MY_INT_MAX;
            else
                row[x] = lut[row[x]];
        }
    }  
}
