/***************************************************************************
                           CFilterMedian.cpp  -  description
                              -------------------
     last modified        : 28.06.2004 Dennis Voss
     copyright            : (C) 2002 Marc Hensel
     email                : tuvisionlib@tuvision.de, marc.hensel@web.de
     origin               : Study thesis Marc Hensel
 
***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#include <tuv2d/CFilterMedian.h>
#include <tuv/CImageByteGray.h>
#include <algorithm>

using namespace std;
using namespace TUVision;


/**************************************************
 *
 * Constructor and destructor
 *
 **************************************************/

/*/
//! Constructor (sets filter width to 3 if called without argument).
/*/

CFilterMedian::CFilterMedian(unsigned char width)
{
    filterWidth = width;
}


/*/
//! Destructor.
/*/

CFilterMedian::~CFilterMedian()
{
}



/**************************************************
 *
 * Parameters:
 *
 * - set filter width
 * - return filter width
 *
 **************************************************/

/*/
//! Set filter width (must be an odd value).
/*/

bool CFilterMedian::setWidth(unsigned char width)
{
    if (width%2!=1) return false;			// Input value is even
    else			filterWidth = width;	// Input value is odd
    return true;
}


/*/
//! Return current filter width.
/*/

unsigned char CFilterMedian::getWidth()
{
    return filterWidth;
}



/**************************************************
 *
 * Filter array (1d):
 *
 * - int
 * - unsigned char
 * - double
 * - float
 *
 **************************************************/

/*/
//! Filter a 1-dimensional array.
/*/

bool CFilterMedian::filter(int *data_array,const int array_length)
{
    int		offset;						//!< Filter width / 2 (rounded down)
    int		*orig_array;				//!< Original data array
    int		*sort_array;				//!< Array of elements to be sorted in each filter step
    int		sort_index;					//!< Position in sort_array
    int		filter_position;			//!< Index of currently filtered entry in array
    int		last_position;				//!< Last entry to be filtered (upper bounds in for-construct)
    int		position;					//!< Index of currently used entry in array

    // Check filter width (must be odd)
    if(!(filterWidth%2)) return false;

    // Create sort array (passed to median method) and array holding copy of original data
    sort_array	= new int[filterWidth];
    orig_array	= new int[array_length];
    for (position=0; position<array_length; position++)
        orig_array[position] = data_array[position];

    // Init counters and positions
    offset			= filterWidth / 2;
    last_position	= array_length - offset;

    // Run through pixels (positions to be filtered)
    for (filter_position=offset; filter_position<last_position; filter_position++)
    {
        // Fill array to be searched for median value
        position = filter_position - offset;
        for (sort_index=0; sort_index<filterWidth; sort_index++)
            sort_array[sort_index] = orig_array[position++];

        // Determine and set median value
        // Sort array "sort_array" using sort() in stl <algorithm>
        sort(sort_array, sort_array+array_length);

        data_array[filter_position] = sort_array[filterWidth/2];
    }

    delete[] orig_array;
    delete[] sort_array;
    return true;
}


// ------------------------------

/*/
//! Filter a 1-dimensional array.
/*/

bool CFilterMedian::filter(unsigned char *data_array,const int array_length)
{
    int				offset;						//!< Filter width / 2 (rounded down)
    unsigned char 	*orig_array;				//!< Original data array
    unsigned char 	*sort_array;				//!< Array of elements to be sorted in each filter step
    int				sort_index;					//!< Position in sort_array
    int				filter_position;			//!< Index of currently filtered entry in array
    int				last_position;				//!< Last entry to be filtered (upper bounds in for-construct)
    int				position;					//!< Index of currently used entry in array

    // Check filter width (must be odd)
    if(!(filterWidth%2)) return false;

    // Create sort array (passed to median method) and array holding copy of original data
    sort_array	= new unsigned char[filterWidth];
    orig_array	= new unsigned char[array_length];
    for (position=0; position<array_length; position++)
        orig_array[position] = data_array[position];

    // Init counters and positions
    offset			= filterWidth / 2;
    last_position	= array_length - offset;

    // Run through pixels (positions to be filtered)
    for (filter_position=offset; filter_position<last_position; filter_position++)
    {
        // Fill array to be searched for median value
        position = filter_position - offset;
        for (sort_index=0; sort_index<filterWidth; sort_index++)
            sort_array[sort_index] = orig_array[position++];

        // Determine and set median value
        // Sort array "sort_array" using sort() in stl <algorithm>
        sort(sort_array, sort_array+array_length);

        data_array[filter_position] = sort_array[filterWidth/2];
    }

    delete[] orig_array;
    delete[] sort_array;
    return true;
}


// ------------------------------

/*/
//! Filter a 1-dimensional array.
/*/

bool CFilterMedian::filter(double *data_array,const int array_length)
{
    int		offset;						//!< Filter width / 2 (rounded down)
    double 	*orig_array;				//!< Original data array
    double	*sort_array;				//!< Array of elements to be sorted in each filter step
    int		sort_index;					//!< Position in sort_array
    int		filter_position;			//!< Index of currently filtered entry in array
    int		last_position;				//!< Last entry to be filtered (upper bounds in for-construct)
    int		position;					//!< Index of currently used entry in array

    // Check filter width (must be odd)
    if(!(filterWidth%2)) return false;

    // Create sort array (passed to median method) and array holding copy of original data
    sort_array	= new double[filterWidth];
    orig_array	= new double[array_length];
    for (position=0; position<array_length; position++)
        orig_array[position] = data_array[position];

    // Init counters and positions
    offset			= filterWidth / 2;
    last_position	= array_length - offset;

    // Run through pixels (positions to be filtered)
    for (filter_position=offset; filter_position<last_position; filter_position++)
    {
        // Fill array to be searched for median value
        position = filter_position - offset;
        for (sort_index=0; sort_index<filterWidth; sort_index++)
            sort_array[sort_index] = orig_array[position++];

        // Determine and set median value
        // Sort array "sort_array" using sort() in stl <algorithm>
        sort(sort_array, sort_array+array_length);

        data_array[filter_position] = sort_array[filterWidth/2];
    }

    delete[] orig_array;
    delete[] sort_array;
    return true;
}


// ------------------------------

/*/
//! Filter a 1-dimensional array.
/*/

bool CFilterMedian::filter(float *data_array,const int array_length)
{
    int		offset;						//!< Filter width / 2 (rounded down)
    float	*orig_array;				//!< Original data array
    float	*sort_array;				//!< Array of elements to be sorted in each filter step
    int		sort_index;					//!< Position in sort_array
    int		filter_position;			//!< Index of currently filtered entry in array
    int		last_position;				//!< Last entry to be filtered (upper bounds in for-construct)
    int		position;					//!< Index of currently used entry in array

    // Check filter width (must be odd)
    if(!(filterWidth%2)) return false;

    // Create sort array (passed to median method) and array holding copy of original data
    sort_array	= new float[filterWidth];
    orig_array	= new float[array_length];
    for (position=0; position<array_length; position++)
        orig_array[position] = data_array[position];

    // Init counters and positions
    offset			= filterWidth / 2;
    last_position	= array_length - offset;

    // Run through pixels (positions to be filtered)
    for (filter_position=offset; filter_position<last_position; filter_position++)
    {
        // Fill array to be searched for median value
        position = filter_position - offset;
        for (sort_index=0; sort_index<filterWidth; sort_index++)
            sort_array[sort_index] = orig_array[position++];

        // Determine and set median value
        // Sort array "sort_array" using sort() in stl <algorithm>
        sort(sort_array, sort_array+array_length);

        data_array[filter_position] = sort_array[filterWidth/2];
    }

    delete[] orig_array;
    delete[] sort_array;
    return true;
}



/**************************************************
 *
 * Filter image (2d):
 *
 **************************************************/

/*/
//! Filter an image (2d filter).
/*/

bool CFilterMedian::filter(CImageByteGray &img)
{
    /*
     *	Meaning of 'offset' (e.g. filterWidth = 7):
     *
     *	offset=3
     *	+ - - - - - + <- middle - offset
     *	|     |     |
     *	|     |     |
     *	| - - P - - | <- middle
     *	|     |     |
     *	|     |     |
     *	+ - - - - - + <- middle + offset
     *
     *	P = pixel to be filtered (filter window centered)
     */

    unsigned char	offset;						//!< Filter width / 2 (rounded down)
    CImageByteGray	*orig_img;					//!< Original image
    int				x_win,y_win;				//!< Relative position in filter window
    unsigned char	*row;						//!< Pointer: First pixel in row (image matrix)
    unsigned char	*sort_array;				//!< Array of window elements to be sorted in each filter step
    int				sort_index;					//!< Position in sort_array
    int				sort_array_length;			//!< Number of entries in each filter window (= number values to be sorted)
    int				upperX,upperY;				//!< Last row/column to be filtered (upper bounds in for-constructs)

    // Check filter width (must be odd)
    if(!(filterWidth%2)) return false;

    // Create sort array (passed to median method) and image holding copy of original data
    sort_array_length	= filterWidth * filterWidth;
    sort_array			= new unsigned char[sort_array_length];
    orig_img			= new CImageByteGray (img);
    img.clearImage();

    // Init counters and positions
    offset			= filterWidth / 2;
    upperX			= orig_img->getW() - offset;
    upperY			= orig_img->getH() - offset;

    // Run through pixels (positions to be filtered)
    for (int y=offset; y<upperY; y++)
    {
        for (int x=offset; x<upperX; x++)
        {
            // Fill array to be searched for median value (filter window)
            sort_index = 0;
            for (y_win=-offset; y_win<=offset; y_win++)
            {
                row = orig_img->getRow(y+y_win);
                for (x_win=-offset; x_win<=offset; x_win++)
                    sort_array[sort_index++] = row[x+x_win];
            }

            // Determine and set median value
            // Sort array "sort_array" using sort() in stl <algorithm>
            sort(sort_array, sort_array+sort_array_length);

            img.setPixel( x, y, sort_array[sort_array_length/2] );				// Find median and store in output image
        }
    }
    delete orig_img;
    delete[] sort_array;
    return true;
}
