/***************************************************************************
                           CFilterSumInversion.cpp  -  description
                              -------------------
     begin                : 22.08.2002
     copyright            : (C) 2002 Marc Hensel
     email                : tuvisionlib@tuvision.de, marc.hensel@web.de
     origin               : Study thesis Marc Hensel
 
***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

/*
 *	Steps taken in filter method:
 *
 *	(1) Sum up pixel values in centered window
 *	(2)	Map so that e.g.
 *	    if all pixels in window = 0   => 255 and
 *		if all pixels in window = 255 => 0
 *  (3) Normalize/stretch so that min=0 and max=255
 *	    (Skipped if all pixels have same value)
 */

#include <tuv2d/CFilterSumInversion.h>
#include <tuv/CImageByteGray.h>

// using namespace std;
using namespace TUVision;


/**************************************************
 *
 * Constructor and destructor
 *
 **************************************************/

CFilterSumInversion::CFilterSumInversion(unsigned char width)
{
    filterWidth = width;
}


CFilterSumInversion::~CFilterSumInversion()
{
}


/**************************************************
 *
 * Public methods
 *
 **************************************************/

/*
 * Get and set parameter: filterWidth
 */

bool CFilterSumInversion::setWidth(unsigned char width)
{
    if (width%2!=1) return false;			// Input value is even
    else			filterWidth = width;	// Input value is odd
    return true;
}

unsigned char CFilterSumInversion::getWidth()
{
    return filterWidth;
}


/*
 * Filter image
 */

void CFilterSumInversion::filter(CImageByteGray &img)
{
    /*
     *	Meaning of 'offset' (e.g. filterWidth = 7):
     *
     *	offset=3
     *	+ - - - - - + <- middle - offset
     *	|     |     |
     *	|     |     |
     *	| - - P - - | <- middle
     *	|     |     |
     *	|     |     |
     *	+ - - - - - + <- middle + offset
     *
     *	P = pixel to be filtered (filter window centered)
     */

    if(!(filterWidth%2)) return;			// Filter length has to be odd

    unsigned char	offset = filterWidth/2;			// See above (drawing)
    int				x_win,y_win;					// Relative position in filter window
    unsigned char	*row;							// Pointer: 1st pixel in row (image matrix)
    unsigned char	value,min_val,max_val;			// Value of filtered pixel
    int				sum;							// Sum of pixels in filter window
    double			max = 255*filterWidth*filterWidth;	// Max. possible sum in filter window
    CImageByteGray	*orig_img;

    orig_img = new CImageByteGray (img);
    img.clearImage();

    /* Filter inverted sum */
    min_val = 255;
    max_val = 0;
    // Run through pixels (positions to be filtered)
    for (int y=offset; y<orig_img->getH()-offset; y++)
    {
        for (int x=offset; x<orig_img->getW()-offset; x++)
        {
            // Evaluate filter window
            sum = 0;
            for (y_win=-offset; y_win<=offset; y_win++)
            {
                row = orig_img->getRow(y+y_win);
                for (x_win=-offset; x_win<=offset; x_win++)
                    sum += row[x+x_win];
            }
            value = (unsigned char)(255.0*(1.0-sum/max));
            img.setPixel(x,y,value,0);
            if (min_val>value) min_val = value;
            if (max_val<value) max_val = value;
        }
    }

    /* Normalize to range [0,255] */
    if (min_val!=max_val)
    {
        double	factor = 255.0/(max_val-min_val);
        for (int y=offset; y<img.getH()-offset; y++)
        {
            row = img.getRow(y);
            for (int x=offset; x<img.getW()-offset; x++)
                row[x] = (unsigned char)((row[x]-min_val) * factor);
        }
    }
    delete orig_img;
}


/*
 * Remove border: Set flags around filtered (!) image borders and
 * remove all border areas with pixel value 255.
 */

void CFilterSumInversion::removeBorder(CImageByteGray &img,unsigned char threshold)
{
    unsigned char	offset = filterWidth/2;		// See above (drawing)
    int				x,y;
    CImageByteGray	*flag_img;					// Flag (=1): Pixel still to examine
    unsigned char	*flag_row,*next,*last;		// Pointer to pixel in flag matrix
    bool			rescan;						// Scan last line again
    unsigned char	*row;						// Pointer to pixel in image matrix

    flag_img = new CImageByteGray (img);
    flag_img->clearImage();

    /* Set 'to examine'-flags around border pixels */
    flag_row = flag_img->getRow(offset);					// Top border
    for (x=offset; x<=flag_img->getW()-offset-1; x++)
        flag_row [x] = 1;
    flag_row = flag_img->getRow(flag_img->getH()-offset-1);	// Bottom border
    for (x=offset; x<=flag_img->getW()-offset-1; x++)
        flag_row [x] = 1;
    for (y=offset; y<=flag_img->getH()-offset-1; y++)		// Left and right border
    {
        flag_row = flag_img->getRow(y);
        flag_row [offset] = 1;
        flag_row [flag_img->getW()-offset-1] = 1;
    }

    /* Examine pixels: Scan from top to bottom */
    for (y=offset; y<=flag_img->getH()-offset-1; y++)
    {
        row		 = img.getRow(y);
        flag_row = flag_img->getRow(y);
        next	 = flag_img->getRow(y+1);
        last	 = flag_img->getRow(y-1);
        rescan	 = false;
        for (x=offset; x<=flag_img->getW()-offset-1; x++)
        {
            if ((flag_row[x]==1)&&(row[x]>=threshold))	// Pixel part of border area
            {
                flag_row[x-1]	= 1;					// Set flags of neighbouring pixels
                flag_row[x+1]	= 1;
                last	[x-1]	= 1;
                last	[x]		= 1;
                last	[x+1]	= 1;
                next	[x-1]	= 1;
                next	[x]		= 1;
                next	[x+1]	= 1;
                row		[x]		= 0;	// Set pixel to zero
                rescan			= true;	// Scan last line again
            }
        }
        if (rescan) y = y-2;		// Scan last line again (also compensate 'y++')
    }
    delete flag_img;
}
