/***************************************************************************
                          CCorrelator.cpp  -  description
                             -------------------
    begin                : 01.01.2001
    copyright            : (C) 2001 by Islam Shdaifat
    email                : tuvisionlib@tuvision.de
    origin               : PhD - Thesis Islam Shdaifat

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#include "tuv2d/CCorrelator.h"
#include "tuv/CImageByteGray.h"
#include "tuv/CRect.h"
#include <cmath>
#include <cassert>

using namespace TUVision;

/**********************************************************/
/*!
 * returnes the average of the image
 * \param  img  the input image
 * \return the  average of the image
 */
/**********************************************************/
double
TUVision::calcAvg(const CImageByteGray& img, const int step)
{
    int avg = 0;

    const unsigned char* pixels;

    int h = img.getH();
    int w = img.getW();
    int i, j;
    for(j=0;j<h;j += step) {
        pixels=img.getRow(j);
        for(i=0;i<w;i += step) 
            avg+= *pixels++;
    }
    return (double)(avg*step)/(w*h);
}

/**********************************************************/
/*!
 * returns the variance of the image of known average
 * \param  img the input image
 * \param  avg average of the image
 * \return the variance of the image
 */
/**********************************************************/
double
TUVision::calcVar(const CImageByteGray& img, const double& avg, const int step)
{
    /*
    static int sqrtab[256+255]; //defined just once using static
    static bool initialized=false; //defined just once using static
    if(!initialized) {
    for(int i=-255;i<256;i++)
    sqrtab[i]=i*i;
    sqrtab+=255;
    initialized=true;
    }
    */
    const unsigned char* pixels;
    double var=0.0;
    double t;
    int h = img.getH();
    int w = img.getW();
    int i, j;
    for(j=0;j<h;j += step) {
        pixels=img.getRow(j);
        for(i=0;i<w;i += step) {
            t=*pixels++-avg;
            var+=t*t;
        }
    }

    return var = (var*step)/ (h*w);
}

/**********************************************************/
/*!
 * returns the correlation between two images, used when the two
 * images are with unknown avrage and variance.
 * \param im1 the first image
 * \param im2 the second image
 *
 * \return the correlation coeffecient between two images
 */
/**********************************************************/
double 
TUVision::correlateImg1Img2(const CImageByteGray& im1,const CImageByteGray& im2,const int step)
{
    double im1avg,im1var;
    im1avg=calcAvg(im1, step);
    im1var=calcVar(im1,im1avg, step);

    return correlateImg1Img2(im1,im2,im1avg,im1var, step);

}

/**********************************************************/
/*!
 * calculate the correlation between two images, used when the one
 * image is with KNOWN average and variance, no need to calculate the
 * average and variance once more
 * \param im1 the first image
 * \param im2 the second image
 * \param im1avg the first image average
 * \param im1var the first image variance
 *
 * \return the correlation coeffecient between two images
 */
/**********************************************************/
double 
TUVision::correlateImg1Img2(const CImageByteGray& im1,
                            const CImageByteGray& im2,
                            const double& im1avg,const double& im1var,
                            const int step)
{
    double im2avg,im2var;
    im2avg=calcAvg(im2, step);
    im2var=calcVar(im2,im2avg,step);

    return correlateImg1Img2(im1,im2,im1avg,im1var,im2avg,im2var,step);
}

/**********************************************************/
/*!
 * calculate the correlation between two images, used when BOTH
 * image have a KNOWN average and variance, i.e. if there's no need to 
 * calculate the average and variance once more
 * \param im1 the first image
 * \param im2 the second image
 * \param im1avg the first image's average
 * \param im1var the first image's variance
 * \param im2avg the second image's average
 * \param im2var the second image's variance
 *
 * \return the correlation coeffecient between two images
 */
/**********************************************************/
double 
TUVision::correlateImg1Img2(const CImageByteGray& im1,
                            const CImageByteGray& im2,
                            const double& im1avg,const double& im1var,
                            const double& im2avg,const double& im2var,
                            const int step)
{
    assert(im1.getW()==im2.getW() && im1.getH()==im2.getH());

    double score=0.0;
    const unsigned char *im1ptr, *im2ptr;

    int corrh = im1.getH();
    int corrw = im1.getW();
    int i,j;
    for(j=0; j<corrh; j += step)
    {
        im1ptr = im1.getRow(j);
        im2ptr = im2.getRow(j);

        for(i=0;i<corrw;i += step) {
            score+=((double)*im1ptr++-im1avg)*((double)*im2ptr++-im2avg);
        }
    }
    // what happens if score==0 and one of the variances is 0? (div. by 0)
    if((im1var==0.0)||(im2var==0.0))
        return(0.0);
    score/=sqrt(im1var*im2var)*corrw*corrh;

    return score;
}


/******************************************************/
/*! returs an image contains the correlation coefficient
 *  by shifting the small image one pixel at a time.
 *  The corr. coeff is mapped from (-1,1) to (0 , 255)
 *  to be fit in the gray byte image. Also the location
 *  of the maximum correlation. This can be used for 
 *  pattern search.
 *
 * \param imLarge the large image to be correlated with
 * \param imSmall image the large image to be correlated
 * \param x_max x of the maximum correlation, returned
 * \param y_max y of the maximum correlation, returned
 * \param corr_max the maximum correlation, returned
 * \param step zoom factor to accelerate the process
 *
 * \return the correlation image
 */
/******************************************************/
CImageByteGray _CCTUVISION 
*TUVision::correlateImgSmallImgLarge(
                                     CImageByteGray& imLarge,
                                     const CImageByteGray& imSmall,
                                     int* x_max, int* y_max, double* corr_max,
                                     int step)
{
    CImageByteGray *im_out =new CImageByteGray(imLarge.getW(),imLarge.getH());
    im_out->clearImage();
    double avg1 = calcAvg(imSmall, step); // the small image average
    double var1 = calcVar(imSmall,avg1, step); // the small image variance

    int x,y;
    int w = imSmall.getW();
    int h = imSmall.getH();

    double  crr;
    *corr_max= -1.0;

    int y_f = imLarge.getH() - imSmall.getH();
    int x_f = imLarge.getW() - imSmall.getW();

    // store the large image window
    CRect rectLarge = imLarge.getWindow();

    // using the window of the large image as offset
    int offsetImLargeX = rectLarge.getX();
    int offsetImLargeY = rectLarge.getY();

    for(y = 0 ; y < y_f ; y += step) // for the image x0-->x_f and y_0-->y_f
    {
        unsigned char* dst = im_out->getRow(y);

        for(x = 0 ; x < x_f; x += step) 
        {
            // the new subimage
            imLarge.setWindow( offsetImLargeX + x, offsetImLargeY + y, w, h ); 

            crr = correlateImg1Img2( imSmall, imLarge, avg1, var1, step );

            if(crr > *corr_max) // maximum correlation
            {
                *corr_max = crr ; *x_max = x; *y_max = y;
            }

            for(int m = 0; m < step; m++)
                if(crr> 0.7)
                    *dst++ = (unsigned char)(crr*128.0 + 127.0);
                else dst++;
                //dst += step;
        }

    }

    imLarge.setWindow( rectLarge );;

    return(im_out);
}


/*!
\anchor correlateImg1Img2Doc

*/

