/***************************************************************************
                          COperatorConverter.cpp  -  description
                             -------------------
    begin                : 01.01.2001
    copyright            : (C) 2001 by Eric Sommerlade
    email                : tuvisionlib@tuvision.de
    origin               : Studienarbeit/Diplomarbeit Eric Sommerlade

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#include "tuv2d/COperatorConverter.h"

#include "tuv/CImageByteGray.h"
#include "tuv/CImageByteRGB.h"
#include "tuv/CImageByteYUV.h"
#include "tuv/CImageIntGray.h"
#include "tuv/CImageIntRGB.h"
#include "tuv/CImageIntYUV.h"
#include "tuv/CImageDoubleGray.h"
#include "tuv/CImageDoubleRGB.h"
#include "tuv/CImageDoubleYUV.h"

#include <climits>


//using namespace std;
using namespace TUVision;



/**********************************************************/
/*! 
 *   Converts a CImageByte<ColorModel> image to a 
 *   CImageInt<ColorModel> image.
 *
 *   \param imgsrc    source image
 *   \param imgdst    destination image
 *   \param nchannels number of color channels
 */
/**********************************************************/

void
COperatorConverter::convertByteToInt (const CImageByte &imgsrc, 
                                      CImageInt &imgdst,
                                      int nchannels)
{
    int                  x, y;
    const unsigned char  *srcpnt;
    int                  *dstpnt;
    int                  width;


    width = imgsrc.getW();

    for (y = 0; y < imgsrc.getH(); y++)
    {
        srcpnt = imgsrc.getRow (y);
        dstpnt = imgdst.getRow (y);

        for (x = 0; x < width * nchannels; x++)
            dstpnt[x] = srcpnt[x] << 8; 
    }
}


/**********************************************************/
/*! 
 *   Converts a CImageByte<ColorModel> image to a 
 *   CImageDouble<ColorModel> image.
 *
 *   \param imgsrc    source image
 *   \param imgdst    destination image
 *   \param nchannels number of color channels
 */
/**********************************************************/

void
COperatorConverter::convertByteToDouble (const CImageByte &imgsrc, 
                                         CImageDouble &imgdst,
                                         int nchannels)
{
    int                  x, y;
    const unsigned char  *srcpnt;
    double               *dstpnt;
    int                  width;


    width = imgsrc.getW();

    for (y = 0; y < imgsrc.getH(); y++)
    {
        srcpnt = imgsrc.getRow (y);
        dstpnt = imgdst.getRow (y);

        for (x = 0; x < width * nchannels; x++)
            dstpnt[x] = ((double) srcpnt[x]) / 255.0; 
    }
}


/**********************************************************/
/*! 
 *   Converts a CImageInt<ColorModel> image to a 
 *   CImageByte<ColorModel> image. Negative values are 
 *   clipped to "0" and values larger than "2^16-1" are 
 *   clipped to "255". 
 *
 *   \param imgsrc source image
 *   \param imgdst destination image
 *   \param nchannels number of color channels
 */
/**********************************************************/

void
COperatorConverter::convertIntToByte (const CImageInt &imgsrc, 
                                      CImageByte &imgdst,
                                      int nchannels)
{
    int            x, y;
    const int      *srcpnt;
    unsigned char  *dstpnt;
    int            width;
    unsigned int   tmp;


    width = imgsrc.getW();

    for (y = 0; y < imgsrc.getH(); y++)
    {
        srcpnt = imgsrc.getRow (y);
        dstpnt = imgdst.getRow (y);

        for (x = 0; x < width * nchannels; x++)
        {
            if (srcpnt[x] < 0)
                dstpnt[x] = 0;
            else 
            {
                tmp = ((unsigned int) srcpnt[x]) >> 8;
                if (tmp > 255)
                    dstpnt[x] = 255;
                else
                    dstpnt[x] = tmp;
            }
        }
    }
}


/**********************************************************/
/*! 
 *   Converts a CImageInt<ColorModel> image to a 
 *   CImageDouble<ColorModel> image. 
 *
 *   \param imgsrc    source image
 *   \param imgdst    destination image
 *   \param nchannels number of color channels
 */
/**********************************************************/

void
COperatorConverter::convertIntToDouble (const CImageInt &imgsrc, 
                                        CImageDouble &imgdst,
                                        int nchannels)
{
    int                  x, y;
    const int            *srcpnt;
    double               *dstpnt;
    int                  width;


    width = imgsrc.getW();

    for (y = 0; y < imgsrc.getH(); y++)
    {
        srcpnt = imgsrc.getRow (y);
        dstpnt = imgdst.getRow (y);

        for (x = 0; x < width * nchannels; x++)
            dstpnt[x] = ((double) srcpnt[x]) / 65535.0; 
    }
}


/**********************************************************/
/*! 
 *   Converts a CImageDouble<ColorModel> image to a 
 *   CImageByte<ColorModel> image. Negative values are 
 *   clipped to "0" and values larger than "1.0" are 
 *   clipped to "255". 
 *
 *   \param imgsrc source image
 *   \param imgdst destination image
 *   \param nchannels number of color channels
 */
/**********************************************************/

void
COperatorConverter::convertDoubleToByte (const CImageDouble &imgsrc, 
                                         CImageByte &imgdst,
                                         int nchannels)
{
    int            x, y;
    const double   *srcpnt;
    unsigned char  *dstpnt;
    int            width;


    width = imgsrc.getW();

    for (y = 0; y < imgsrc.getH(); y++)
    {
        srcpnt = imgsrc.getRow (y);
        dstpnt = imgdst.getRow (y);

        for (x = 0; x < width * nchannels; x++)
        {
            if (srcpnt[x] < 0.0)
                dstpnt[x] = 0;
            else 
            {
                if (srcpnt[x] > 1.0)
                    dstpnt[x] = 255;
                else
                    dstpnt[x] = (unsigned char) (srcpnt[x] * 255.0);
            }
        }
    }
}


/**********************************************************/
/*! 
 *   Converts a CImageDouble<ColorModel> image to a 
 *   CImageInt<ColorModel> image. Values smaller or
 *   larger than INT_MIN/INT_MAX are mapped to
 *   INT_MIN/INT_MAX.
 *
 *   \param imgsrc source image
 *   \param imgdst destination image
 *   \param nchannels number of color channels
 */
/**********************************************************/

void
COperatorConverter::convertDoubleToInt (const CImageDouble &imgsrc, 
                                        CImageInt &imgdst,
                                        int nchannels)
{
    int            x, y;
    const double   *srcpnt;
    int            *dstpnt;
    int            width;


    width = imgsrc.getW();

    for (y = 0; y < imgsrc.getH(); y++)
    {
        srcpnt = imgsrc.getRow (y);
        dstpnt = imgdst.getRow (y);

        for (x = 0; x < width * nchannels; x++)
        {
            if (srcpnt[x] > INT_MAX)
                dstpnt[x] = INT_MAX;
            else if (srcpnt[x] < INT_MIN)
                dstpnt[x] = INT_MIN;
            else 
                dstpnt[x] = (int) (srcpnt[x] * 65535.0);
        }
    }
}

////////////////////////////////////////////////////////////
// Converters for gray images
////////////////////////////////////////////////////////////


void 
COperatorConverter::convert (const CImageByteGray &imgsrc, 
                             CImageIntGray **imgdst)
{
    *imgdst = new CImageIntGray (imgsrc.getW(), imgsrc.getH());
    convertByteToInt (imgsrc, **imgdst, 1);
}


void 
COperatorConverter::convert (const CImageByteGray &imgsrc, 
                             CImageDoubleGray **imgdst)
{
    *imgdst = new CImageDoubleGray (imgsrc.getW(), imgsrc.getH());
    convertByteToDouble (imgsrc, **imgdst, 1);
}


////////////////////////////////////////////////////////////

void 
COperatorConverter::convert (const CImageIntGray &imgsrc, 
                             CImageByteGray **imgdst)
{
    *imgdst = new CImageByteGray (imgsrc.getW(), imgsrc.getH());
    convertIntToByte (imgsrc, **imgdst, 1);
}


void 
COperatorConverter::convert (const CImageIntGray &imgsrc, 
                             CImageDoubleGray **imgdst)
{
    *imgdst = new CImageDoubleGray (imgsrc.getW(), imgsrc.getH());
    convertIntToDouble (imgsrc, **imgdst, 1);
}

////////////////////////////////////////////////////////////

void 
COperatorConverter::convert (const CImageDoubleGray &imgsrc, 
                             CImageByteGray **imgdst)
{
    *imgdst = new CImageByteGray (imgsrc.getW(), imgsrc.getH());
    convertDoubleToByte (imgsrc, **imgdst, 1);
}


void 
COperatorConverter::convert (const CImageDoubleGray &imgsrc, 
                             CImageIntGray **imgdst)
{
    *imgdst = new CImageIntGray (imgsrc.getW(), imgsrc.getH());
    convertDoubleToInt (imgsrc, **imgdst, 1);
}

////////////////////////////////////////////////////////////
// Converters for RGB images
////////////////////////////////////////////////////////////

void 
COperatorConverter::convert (const CImageByteRGB &imgsrc, 
                             CImageIntRGB **imgdst)
{
    *imgdst = new CImageIntRGB (imgsrc.getW(), imgsrc.getH());
    convertByteToInt (imgsrc, **imgdst, 3);
}


void 
COperatorConverter::convert (const CImageByteRGB &imgsrc, 
                             CImageDoubleRGB **imgdst)
{
    *imgdst = new CImageDoubleRGB (imgsrc.getW(), imgsrc.getH());
    convertByteToDouble (imgsrc, **imgdst, 3);
}


////////////////////////////////////////////////////////////

void 
COperatorConverter::convert (const CImageIntRGB &imgsrc, 
                             CImageByteRGB **imgdst)
{
    *imgdst = new CImageByteRGB (imgsrc.getW(), imgsrc.getH());
    convertIntToByte (imgsrc, **imgdst, 3);
}


void 
COperatorConverter::convert (const CImageIntRGB &imgsrc, 
                             CImageDoubleRGB **imgdst)
{
    *imgdst = new CImageDoubleRGB (imgsrc.getW(), imgsrc.getH());
    convertIntToDouble (imgsrc, **imgdst, 3);
}

////////////////////////////////////////////////////////////

void 
COperatorConverter::convert (const CImageDoubleRGB &imgsrc, 
                             CImageByteRGB **imgdst)
{
    *imgdst = new CImageByteRGB (imgsrc.getW(), imgsrc.getH());
    convertDoubleToByte (imgsrc, **imgdst, 3);
}


void 
COperatorConverter::convert (const CImageDoubleRGB &imgsrc, 
                             CImageIntRGB **imgdst)
{
    *imgdst = new CImageIntRGB (imgsrc.getW(), imgsrc.getH());
    convertDoubleToInt (imgsrc, **imgdst, 3);
}



////////////////////////////////////////////////////////////
// Converters for YUV images
////////////////////////////////////////////////////////////

void 
COperatorConverter::convert (const CImageByteYUV &imgsrc, 
                             CImageIntYUV **imgdst)
{
    *imgdst = new CImageIntYUV (imgsrc.getW(), imgsrc.getH());
    convertByteToInt (imgsrc, **imgdst, 3);
}


void 
COperatorConverter::convert (const CImageByteYUV &imgsrc, 
                             CImageDoubleYUV **imgdst)
{
    *imgdst = new CImageDoubleYUV (imgsrc.getW(), imgsrc.getH());
    convertByteToDouble (imgsrc, **imgdst, 3);
}


////////////////////////////////////////////////////////////

void 
COperatorConverter::convert (const CImageIntYUV &imgsrc, 
                             CImageByteYUV **imgdst)
{
    *imgdst = new CImageByteYUV (imgsrc.getW(), imgsrc.getH());
    convertIntToByte (imgsrc, **imgdst, 3);
}


void 
COperatorConverter::convert (const CImageIntYUV &imgsrc, 
                             CImageDoubleYUV **imgdst)
{
    *imgdst = new CImageDoubleYUV (imgsrc.getW(), imgsrc.getH());
    convertIntToDouble (imgsrc, **imgdst, 3);
}

////////////////////////////////////////////////////////////

void 
COperatorConverter::convert (const CImageDoubleYUV &imgsrc, 
                             CImageByteYUV **imgdst)
{
    *imgdst = new CImageByteYUV (imgsrc.getW(), imgsrc.getH());
    convertDoubleToByte (imgsrc, **imgdst, 3);
}


void 
COperatorConverter::convert (const CImageDoubleYUV &imgsrc, 
                             CImageIntYUV **imgdst)
{
    *imgdst = new CImageIntYUV (imgsrc.getW(), imgsrc.getH());
    convertDoubleToInt (imgsrc, **imgdst, 3);
}


//RGB to LUV conversion
const double XYZconv[3][3] = {	{  0.4125,  0.3576,  0.1804 },
{  0.2125,  0.7154,  0.0721 },
{  0.0193,  0.1192,  0.9502 }	};

//LUV to RGB conversion
const double RGBconv[3][3] = {	{  3.2405, -1.5371, -0.4985 },
{ -0.9693,  1.8760,  0.0416 },
{  0.0556, -0.2040,  1.0573 }	};

const double Yn			= 1.00000;
const double Un_prime	= 0.19784977571475;
const double Vn_prime	= 0.46834507665248;
const double Lt			= 0.008856;
#include <cmath>

void COperatorConverter::RGBtoLUV(const unsigned char *rgbVal, double *luvVal)
{

    //delcare variables
    double	x, y, z, L0, u_prime, v_prime, constant;

    //convert RGB to XYZ...
    x		= XYZconv[0][0]*rgbVal[0] + XYZconv[0][1]*rgbVal[1] + XYZconv[0][2]*rgbVal[2];
    y		= XYZconv[1][0]*rgbVal[0] + XYZconv[1][1]*rgbVal[1] + XYZconv[1][2]*rgbVal[2];
    z		= XYZconv[2][0]*rgbVal[0] + XYZconv[2][1]*rgbVal[1] + XYZconv[2][2]*rgbVal[2];

    //convert XYZ to LUV...

    //compute L*
    L0		= y / (255.0 * Yn);
    if(L0 > Lt)
        luvVal[0]	= (double)(116.0 * (pow(L0, 1.0/3.0)) - 16.0);
    else
        luvVal[0]	= (double)(903.3 * L0);

    //compute u_prime and v_prime
    constant	= x + 15 * y + 3 * z;
    if(constant != 0)
    {
        u_prime	= (4 * x) / constant;
        v_prime = (9 * y) / constant;
    }
    else
    {
        u_prime	= 4.0;
        v_prime	= 9.0/15.0;
    }

    //compute u* and v*
    luvVal[1] = (double) (13 * luvVal[0] * (u_prime - Un_prime));
    luvVal[2] = (double) (13 * luvVal[0] * (v_prime - Vn_prime));

    //done.
    return;

}

inline int my_round(double in_x)
{
    if (in_x < 0)
        return (int)(in_x - 0.5);
    else
        return (int)(in_x + 0.5);
}

void COperatorConverter::LUVtoRGB(const double *luvVal, unsigned char *rgbVal)
{

    //declare variables...
    int		r, g, b;
    double	x, y, z, u_prime, v_prime;

    //perform conversion
    if(luvVal[0] < 0.1)
        r = g = b = 0;
    else
    {
        //convert luv to xyz...
        if(luvVal[0] < 8.0)
            y	= Yn * luvVal[0] / 903.3;
        else
        {
            y	= (luvVal[0] + 16.0) / 116.0;
            y  *= Yn * y * y;
        }

        u_prime	= luvVal[1] / (13 * luvVal[0]) + Un_prime;
        v_prime	= luvVal[2] / (13 * luvVal[0]) + Vn_prime;

        x		= 9 * u_prime * y / (4 * v_prime);
        z		= (12 - 3 * u_prime - 20 * v_prime) * y / (4 * v_prime);

        //convert xyz to rgb...
        //[r, g, b] = RGBconv*[x, y, z]*255.0
        r		= my_round((RGBconv[0][0]*x + RGBconv[0][1]*y + RGBconv[0][2]*z)*255.0);
        g		= my_round((RGBconv[1][0]*x + RGBconv[1][1]*y + RGBconv[1][2]*z)*255.0);
        b		= my_round((RGBconv[2][0]*x + RGBconv[2][1]*y + RGBconv[2][2]*z)*255.0);

        //check bounds...
        if(r < 0)	r = 0; if(r > 255)	r = 255;
        if(g < 0)	g = 0; if(g > 255)	g = 255;
        if(b < 0)	b = 0; if(b > 255)	b = 255;

    }

    //assign rgb values to rgb vector rgbVal
    rgbVal[0]	= r;
    rgbVal[1]	= g;
    rgbVal[2]	= b;

    //done.
    return;
}

void
COperatorConverter::convert (const CImageByteRGB &imgsrc, 
                             CImageDoubleYUV **imgdst)
{
    *imgdst = new CImageDoubleYUV (imgsrc.getW(), imgsrc.getH());

    int                  x, y;
    const unsigned char  *srcpnt;
    double               *dstpnt;
    int                  width;


    width = imgsrc.getW();

    for (y = 0; y < imgsrc.getH(); y++)
    {
        srcpnt = imgsrc.getRow (y);
        dstpnt = (*imgdst)->getRow (y);

        for (x = 0; x < width ; x++)
        {
            RGBtoLUV(srcpnt, dstpnt);
            srcpnt += 3;
            dstpnt += 3;
        }
    }

}

void
COperatorConverter::convert (const CImageDoubleYUV &imgsrc, 
                             CImageByteRGB **imgdst)
{
    *imgdst = new CImageByteRGB (imgsrc.getW(), imgsrc.getH());

    int                   x, y;
    const double        *srcpnt;
    unsigned char       *dstpnt;
    int                   width;


    width = imgsrc.getW();

    for (y = 0; y < imgsrc.getH(); y++)
    {
        srcpnt = imgsrc.getRow (y);
        dstpnt = (*imgdst)->getRow (y);

        for (x = 0; x < width ; x++)
        {
            LUVtoRGB(srcpnt, dstpnt);
            srcpnt += 3;
            dstpnt += 3;
        }
    }

}


void
COperatorConverter::convertRGBtoYUV (const CImageByteRGB &imgsrc, 
                                     CImageDoubleYUV **imgdst)
{
    *imgdst = new CImageDoubleYUV (imgsrc.getW(), imgsrc.getH());

    int                  x, y;
    const unsigned char  *srcpnt;
    double               *dstpnt;
    int                  width;


    width = imgsrc.getW();

    for (y = 0; y < imgsrc.getH(); y++)
    {
        srcpnt = imgsrc.getRow (y);
        dstpnt = (*imgdst)->getRow (y);

        for (x = 0; x < width ; x++)
        {
            RGBtoLUV(srcpnt, dstpnt);
            srcpnt += 3;
            dstpnt += 3;
        }
    }

}

void
COperatorConverter::convertYUVtoRGB (const CImageDoubleYUV &imgsrc, 
                                     CImageByteRGB **imgdst)
{
    *imgdst = new CImageByteRGB (imgsrc.getW(), imgsrc.getH());

    int                   x, y;
    const double        *srcpnt;
    unsigned char       *dstpnt;
    int                   width;


    width = imgsrc.getW();

    for (y = 0; y < imgsrc.getH(); y++)
    {
        srcpnt = imgsrc.getRow (y);
        dstpnt = (*imgdst)->getRow (y);

        for (x = 0; x < width ; x++)
        {
            LUVtoRGB(srcpnt, dstpnt);
            srcpnt += 3;
            dstpnt += 3;
        }
    }

}
