/***************************************************************************
                          COperatorHistogram.cpp  -  description
                             -------------------
    begin                : 01.01.2001
    copyright            : (C) 2001 by Islam Shdaifat
    email                : tuvisionlib@tuvision.de
    origin               : PhD - Thesis Islam Shdaifat

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#include "tuv2d/COperatorHistogram.h"
#include "tuv/CImageByteGray.h"
#include "tuv/CImageIntGray.h"
//#include "tuv2d/COperatorDraw.h"


//using namespace std;
using namespace TUVision;


/**********************************************************/
/*!
 *   Constructor.
 */
/**********************************************************/

COperatorHistogram::COperatorHistogram()
{
    d_histogram_size = 0;
    d_histogram_data = 0;
}


/**********************************************************/
/*!
 *   Destructor.
 */
/**********************************************************/

COperatorHistogram::~COperatorHistogram()
{
    delete [] d_histogram_data;
}


/**********************************************************/
/*!
 *   Calculate the histogram for the given image object
 *   (byte gray). The calculated histogram is stored in
 *   this class.
 *
 *   \param img image object
 */
/**********************************************************/

void COperatorHistogram::setImage (const CImageByteGray &img)
{
    int                  x, y;
    const unsigned char  *row;



    d_histogram_size = 256;

    // Get buffer for the histogram.
    delete [] d_histogram_data;
    d_histogram_data = new int [d_histogram_size];
    for (x = 0; x < d_histogram_size; x++)
        d_histogram_data[x] = 0;

    // Calculate histogram.
    for (y = 0; y < img.getH(); y++)
    {
        row = img.getRow (y);

        for (x = 0; x < img.getW(); x++)
            d_histogram_data[row[x]]++;
    }
}


/**********************************************************/
/*!
 *   Calculate the histogram for the given image object
 *   (integer gray). The calculated histogram is stored in
 *   this class.
 *
 *   \param img image object
 */
/**********************************************************/

void COperatorHistogram::setImage (const CImageIntGray &img)
{
    int         x, y;
    const int   *row;



    d_histogram_size = 65536;

    // Get buffer for the histogram.
    delete [] d_histogram_data;
    d_histogram_data = new int [d_histogram_size];
    for (x = 0; x < d_histogram_size; x++)
        d_histogram_data[x] = 0;

    // Calculate histogram.
    for (y = 0; y < img.getH(); y++)
    {
        row = img.getRow (y);

        for (x = 0; x < img.getW(); x++)
            if ((row[x] >= 0) && (row[x] < d_histogram_size))
                d_histogram_data[row[x]]++;
    }
}


/**********************************************************/
/*!
 *   Get a const pointer to histogram calculated by the
 *   setImage() function. The histogram consists of a
 *   int array of size getHistogramSize().
 *
 *   \return pointer to the histogram data
 */
/**********************************************************/

const int 
*COperatorHistogram::getHistogramData (void) const
{
    return (d_histogram_data);
}


/**********************************************************/
/*!
 *   Get the number of elements in the calculated
 *   histogram array, which can be accessed by
 *   getHistogramData().
 *
 *   \return number of elements
 */ 
/**********************************************************/

int 
COperatorHistogram::getHistogramSize (void) const
{
    return (d_histogram_size);
}


/**********************************************************/
/*!  Get max index of a nonzero value.
 */ 
/**********************************************************/
int
COperatorHistogram::getMaxNonZero(void) const
{
    for (int i = d_histogram_size-1; i >= 0; --i) {
        if (d_histogram_data[i] != 0)
            return i;
    }
    return 0;
}


/**********************************************************/
/*!
 *   Search the black point of an image in its histogram.
 *   The number of pixels with colors from 0 to the
 *   return color value is larger/equal than the
 *   threshold value.
 *
 *   \return the black point
 */
/**********************************************************/

int 
COperatorHistogram::getBlackpoint (int threshold) const
{
    int x;
    int accu = 0;


    if (threshold < 0)
        return (-1);

    for (x = 0; x < d_histogram_size; x++)
    {
        accu += d_histogram_data[x];
        if (accu >= threshold)
            return (x);
    }

    return (-1);
}


/**********************************************************/
/*!
 *   Search the white point of an image in its histogram.
 *   The number of pixels with colors from the return color 
 *   value to MAX is larger/equal than the threshold value.
 *
 *   \return the white point
 */
/**********************************************************/

int 
COperatorHistogram::getWhitepoint (int threshold) const
{
    int x;
    int accu = 0;


    if (threshold < 0)
        return (-1);

    for (x = d_histogram_size - 1; x > 0; x--)
    {
        accu += d_histogram_data[x];
        if (accu >= threshold)
            return (x);
    }

    return (-1);
}

/*
 * Equalize the image img1 to be the same as img2
 * If img2 = 0 the image will be equalize from its ohn histogram
 *
 * \param im1 the image to be equalized
 * \param im2 the image that contains the equalization histogram data
 */
void
COperatorHistogram::equalizeImage(CImageByteGray *img1, CImageByteGray *img2)
{

    unsigned char
        *equalize_map;
    const unsigned int MaxRGB = 255;
    register int
        i, j, x, y;

    int
        high,
        *histogram,
        low,
        *map;

    /*
    Allocate and initialize histogram arrays.
    */
    if(img2 == 0)
        setImage (*img1);
    else
        setImage (*img2);
    histogram = d_histogram_data;
    map = new int[MaxRGB +1];
    equalize_map =  new unsigned char[MaxRGB +1];


    /*
    Integrate the histogram to get the equalization map.
    */
    j=0;
    for (i=0; i <= MaxRGB; i++)
    {
        j+=histogram[i];
        map[i]=j;
    }
    //delete histogram;
    if (map[MaxRGB] == 0)
    {
        delete equalize_map;
        delete map;
        return;
    }
    /*
    Equalize.
    */
    low=map[0];
    high=map[MaxRGB];
    double high_low = 1;
    if( (high - low) > 1) high_low = (double)(MaxRGB)/(double)(map[MaxRGB] - map[0]);

    for (i=0; i <= MaxRGB; i++)
        equalize_map[i]=(unsigned char)( (double)(map[i]-low)* high_low);

    delete map;

    /*
    Stretch the histogram.
    */
    unsigned char  *row;

    for (y = 0; y < img1->getH(); y++)
    {
        row = img1->getRow (y);

        for (x = 0; x < img1->getW(); x++)
        {
            (*row) = equalize_map[(*row)];
            row++;
        }
    }

    delete equalize_map;
}


/*
 * Create an image from the histogram-data (black on white)
 *
 * \param min		the startpoint of the histgoram (min >= 0)
 * \param max		the endpoint of the histogram (max > min, max < size)
 * \param scale		set true if hist-values should be scaled so the result image
 *                  is square. (This is for better view of the histogram, since
 *                  images with dimensions e.g. 8000x256 are not very good to view)
 */
CImageByteGray*
COperatorHistogram::getHistogramImage(const int min, const int max, const bool scale)
{
    if (min < 0 ||
        max <= min ||
        max > d_histogram_size)
    {
        return 0;
    }
    int height = max - min + 1; // min and max inclusive
    // find max hist-value
	int max_val(0);
	for (int i = min; i <= max; ++i)
		max_val = std::max(max_val, d_histogram_data[i]);

    int width(max_val + 1); // from 0 to max_val
	double width_mul(1.0);
	if (scale) {
		width = height; // square image
		width_mul = double(width) / (max_val + 1); // scale-factor
	}

    // create result-image
    CImageByteGray* result = new CImageByteGray(width, height);
	// clear image (set to white)
	TUVision::CImageByteGray::PixelType* it = result->begin();
	for (; it != result->end(); ++it)
		*it = 255;

	// draw hist-values
    unsigned char  *row;
    for (int y = 0; y < height; ++y) {
        row = result->getRow(y);
		int x_val = int(width_mul*d_histogram_data[y+min]); // scale value for better view
		for (int x = 0; x < x_val; ++x)
			*row++ = 0;
    }

    return result;
}
