/***************************************************************************
                          COperatorScaling.cpp  -  description
                             -------------------
    begin                : 01.01.2001
    copyright            : (C) 2001 by Islam Shdaifat
    email                : tuvisionlib@tuvision.de
    origin               : PhD - Thesis Islam Shdaifat

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
#include "tuv2d/COperatorScaling.h"
#include "tuv/CImageByte.h"
#include "tuv/CImageByteGray.h"
#include "tuv/CImageByteRGB.h"
#include "tuv/CImageByteYUV.h"
#include "tuv/CImageInt.h"
#include "tuv/CImageIntGray.h"
#include "tuv/CImageIntRGB.h"
#include "tuv/CImageIntYUV.h"
#include "tuv/CImageDouble.h"
#include "tuv/CImageDoubleGray.h"
#include "tuv/CImageDoubleRGB.h"
#include "tuv/CImageDoubleYUV.h"
#include "tuv/CImageByteBilevel.h"


//using namespace std;
using namespace TUVision;


/**********************************************************/
/*!
 *   Constructor
 */
/**********************************************************/

COperatorScaling::COperatorScaling(double scaleFactor )
{
    d_scaling = scaleFactor;
}


/**********************************************************/
/*!
 *   Sets the scaling factor to be used.
 *
 *   \param scaling scaling factor (> 0.0)
 */
/**********************************************************/

void COperatorScaling::setScalingFactor (double scalingFactor, double scalingFactorY)
{
    if (scalingFactorY == -1)
    {
        scalingFactorY = scalingFactor;
    }
    if (scalingFactorY < 0.0)
    {
        scalingFactorY = scalingFactor;
    }

    if (scalingFactor > 0.0)
    {
        d_scaling = scalingFactor;
        d_scalingY = scalingFactorY;
    }
}


////////////////////////////////////////////////////////////
//  generic scaling code
////////////////////////////////////////////////////////////

/**********************************************************/
/*!
 *   Uses the given input image to calculate an output
 *   image, which dimensions are "scaling factor" times
 *   smaller than that of the input image (byte version).
 *
 *   \param img_in input image
 *   \param img_out output image
 *   \param number of color channels used by the images
 */
/**********************************************************/

void 
COperatorScaling::scalingGeneric (const CImageByte &img_in, 
                                  CImageByte &img_out, 
                                  int nchannels) const
{
    int   x, y;         // counter "(x,y) in the output image"
    int   x_rel, y_rel; // counter "(dx,dy) of pixel to be combined"
    int   tmp;          // current pixel value
    int   ch;           // counter "color channel"

    int h_out  = img_out.getH(); // hieght of the output image
    int w_out  = img_out.getW(); // width of the output image

    int h_in  = img_in.getH(); // hieght of the input image
    int w_in  = img_in.getW(); // width of the inout image

    for (ch = 0; ch < nchannels; ch++)
    {
        for (y = 0; y < h_out; y++)
        {
            for (x = 0; x < w_out; x++)
            {
                tmp = img_in.getPixel ( (x * w_in)/w_out ,
                    (y * h_in)/h_out , ch);      
                img_out.setPixel (x, y, tmp, ch);
            }
        }
    }
}


/**********************************************************/
/*!
 *   Uses the given input image to calculate an output
 *   image, which dimensions are "scaling factor" times
 *   smaller than that of the input image (integer version).
 *
 *   \param img_in input image
 *   \param img_out output image
 *   \param number of color channels used by the images
 */
/**********************************************************/

void 
COperatorScaling::scalingGeneric (const CImageInt &img_in, 
                                  CImageInt &img_out, 
                                  int nchannels) const
{
    int   x, y;         // counter "(x,y) in the output image"
    int   x_rel, y_rel; // counter "(dx,dy) of pixel to be combined"
    int   tmp;          // current pixel value
    int   ch;           // counter "color channel"

    int h_out  = img_out.getH(); // hieght of the output image
    int w_out  = img_out.getW(); // width of the output image

    int h_in  = img_in.getH(); // hieght of the input image
    int w_in  = img_in.getW(); // width of the inout image

    for (ch = 0; ch < nchannels; ch++)
    {
        for (y = 0; y < h_out; y++)
        {
            for (x = 0; x < w_out; x++)
            {
                tmp = img_in.getPixel ( (x * w_in)/w_out ,
                    (y * h_in)/h_out , ch);      
                img_out.setPixel (x, y, tmp, ch);
            }
        }
    }
}


/**********************************************************/
/*!
 *   Uses the given input image to calculate an output
 *   image, which dimensions are "scaling factor" times
 *   smaller than that of the input image (double version).
 *
 *   \param img_in input image
 *   \param img_out output image
 *   \param number of color channels used by the images
 */
/**********************************************************/

void 
COperatorScaling::scalingGeneric (const CImageDouble &img_in, 
                                  CImageDouble &img_out, 
                                  int nchannels) const
{
    int     ch;           // counter "color channel"
    int   x, y;         // counter "(x,y) in the output image"
    int   x_rel, y_rel; // counter "(dx,dy) of pixel to be combined"
    double  tmp;          // current pixel value

    int h_out  = img_out.getH(); // hieght of the output image
    int w_out  = img_out.getW(); // width of the output image

    int h_in  = img_in.getH(); // hieght of the input image
    int w_in  = img_in.getW(); // width of the inout image

    for (ch = 0; ch < nchannels; ch++)
    {
        for (y = 0; y < h_out; y++)
        {
            for (x = 0; x < w_out; x++)
            {
                tmp = img_in.getPixel ( (x * w_in)/w_out ,
                    (y * h_in)/h_out , ch);      
                img_out.setPixel (x, y, tmp, ch);
            }
        }
    }
}


////////////////////////////////////////////////////////////
//  byte based public functions
////////////////////////////////////////////////////////////

/**********************************************************/
/*!
 *   Uses the given input image to calculate an output
 *   image, which dimensions are "scaling factor" times
 *   smaller than that of the input image (byte gray).
 *   The user is responsible for deleting the returned
 *   image object.
 *
 *   \param img    input image
 *   \param return scaled image
 */
/**********************************************************/

CImageByteGray 
*COperatorScaling::scale (const CImageByteGray &img) const
{
    CImageByteGray  *my_img;


    my_img = new CImageByteGray (img.getW() * d_scaling ,
        img.getH() * d_scalingY);

    scalingGeneric (img, *my_img, 1);

    return (my_img);
}


/**********************************************************/
/*!
 *   Uses the given input image to calculate an output
 *   image, which dimensions are "scaling factor" times
 *   smaller than that of the input image (byte RGB).
 *   The user is responsible for deleting the returned
 *   image object.
 *
 *   \param img    input image
 *   \param return scaled image
 */
/**********************************************************/

CImageByteRGB
*COperatorScaling::scale (const CImageByteRGB &img) const
{
    CImageByteRGB  *my_img;


    my_img = new CImageByteRGB (img.getW() * d_scaling ,
        img.getH() * d_scaling);

    scalingGeneric (img, *my_img, 3);

    return (my_img);
}


/**********************************************************/
/*!
 *   Uses the given input image to calculate an output
 *   image, which dimensions are "scaling factor" times
 *   smaller than that of the input image (byte YUV).
 *   The user is responsible for deleting the returned
 *   image object.
 *
 *   \param img    input image
 *   \param return scaled image
 */
/**********************************************************/

CImageByteYUV
*COperatorScaling::scale (const CImageByteYUV &img) const
{
    CImageByteYUV  *my_img;


    my_img = new CImageByteYUV (img.getW() * d_scaling ,
        img.getH() * d_scaling);

    scalingGeneric (img, *my_img, 3);

    return (my_img);
}



////////////////////////////////////////////////////////////
//  integer based public functions
////////////////////////////////////////////////////////////

/**********************************************************/
/*!
 *   Uses the given input image to calculate an output
 *   image, which dimensions are "scaling factor" times
 *   smaller than that of the input image (integer gray).
 *   The user is responsible for deleting the returned
 *   image object.
 *
 *   \param img    input image
 *   \param return scaled image
 */
/**********************************************************/

CImageIntGray 
*COperatorScaling::scale (const CImageIntGray &img) const
{
    CImageIntGray  *my_img;


    my_img = new CImageIntGray (img.getW() * d_scaling ,
        img.getH() * d_scaling);

    scalingGeneric (img, *my_img, 1);

    return (my_img);
}


/**********************************************************/
/*!
 *   Uses the given input image to calculate an output
 *   image, which dimensions are "scaling factor" times
 *   smaller than that of the input image (integer RGB).
 *   The user is responsible for deleting the returned
 *   image object.
 *
 *   \param img    input image
 *   \param return scaled image
 */
/**********************************************************/

CImageIntRGB
*COperatorScaling::scale (const CImageIntRGB &img) const
{
    CImageIntRGB  *my_img;


    my_img = new CImageIntRGB (img.getW() * d_scaling ,
        img.getH() * d_scaling);

    scalingGeneric (img, *my_img, 3);

    return (my_img);
}


/**********************************************************/
/*!
 *   Uses the given input image to calculate an output
 *   image, which dimensions are "scaling factor" times
 *   smaller than that of the input image (integer YUV).
 *   The user is responsible for deleting the returned
 *   image object.
 *
 *   \param img    input image
 *   \param return scaled image
 */
/**********************************************************/

CImageIntYUV
*COperatorScaling::scale (const CImageIntYUV &img) const
{
    CImageIntYUV  *my_img;


    my_img = new CImageIntYUV (img.getW() * d_scaling ,
        img.getH() * d_scaling);

    scalingGeneric (img, *my_img, 3);

    return (my_img);
}



////////////////////////////////////////////////////////////
//  double based public functions
////////////////////////////////////////////////////////////

/**********************************************************/
/*!
 *   Uses the given input image to calculate an output
 *   image, which dimensions are "scaling factor" times
 *   smaller than that of the input image (double gray).
 *   The user is responsible for deleting the returned
 *   image object.
 *
 *   \param img    input image
 *   \param return scaled image
 */
/**********************************************************/

CImageDoubleGray 
*COperatorScaling::scale (const CImageDoubleGray &img) const
{
    CImageDoubleGray  *my_img;


    my_img = new CImageDoubleGray (img.getW() * d_scaling ,
        img.getH() * d_scaling);

    scalingGeneric (img, *my_img, 1);

    return (my_img);
}


/**********************************************************/
/*!
 *   Uses the given input image to calculate an output
 *   image, which dimensions are "scaling factor" times
 *   smaller than that of the input image (double RGB).
 *   The user is responsible for deleting the returned
 *   image object.
 *
 *   \param img    input image
 *   \param return scaled image
 */
/**********************************************************/

CImageDoubleRGB
*COperatorScaling::scale (const CImageDoubleRGB &img) const
{
    CImageDoubleRGB  *my_img;


    my_img = new CImageDoubleRGB (img.getW() * d_scaling ,
        img.getH() * d_scaling);

    scalingGeneric (img, *my_img, 3);

    return (my_img);
}


/**********************************************************/
/*!
 *   Uses the given input image to calculate an output
 *   image, which dimensions are "scaling factor" times
 *   smaller than that of the input image (byte YUV).
 *   The user is responsible for deleting the returned
 *   image object.
 *
 *   \param img    input image
 *   \param return scaled image
 */
/**********************************************************/

CImageDoubleYUV
*COperatorScaling::scale (const CImageDoubleYUV &img) const
{
    CImageDoubleYUV  *my_img;


    my_img = new CImageDoubleYUV (img.getW() * d_scaling ,
        img.getH() * d_scaling);

    scalingGeneric (img, *my_img, 3);

    return (my_img);
}


CImageByteBilevel
*COperatorScaling::scale (const CImageByteBilevel &img_in) const
{

    int   x, y;         // counter "(x,y) in the output image"
    int   x_rel, y_rel; // counter "(dx,dy) of pixel to be combined"
    int   tmp;          // current pixel value
    int   ch = 0;           // counter "color channel"

    CImageByteBilevel *img_out = new CImageByteBilevel (
        img_in.getW() * d_scaling,
        img_in.getH() * d_scaling);

    int h_out  = img_out->getH(); // hieght of the output image
    int w_out  = img_out->getW(); // width of the output image

    int h_in  = img_in.getH(); // hieght of the input image
    int w_in  = img_in.getW(); // width of the inout image

    for (y = 0; y < h_out; y++)
    {
        for (x = 0; x < w_out; x++)
        {
            tmp = img_in.getPixel ( (x * w_in)/w_out ,
                (y * h_in)/h_out , ch);      
            img_out->setPixel (x, y, tmp, ch);
        }
    }

    return (img_out);

}
