/***************************************************************************
                          OpticalFlow.cpp  -  description
                             -------------------
    begin                : 01.01.2001
    copyright            : (C) 2001 by Eric Sommerlade
    email                : tuvisionlib@tuvision.de
    origin               : Studienarbeit/Diplomarbeit Eric Sommerlade

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#include "tuv/CImageIntGray.h"
#include "tuv/CImageDoubleGray.h"
#include "tuv/CImageByteGray.h"
#include "tuv/CImageByteRGB.h"
#include "tuv2d/OpticalFlow.h"

#include "tuv2d/COperatorGray.h"
#include "tuv2d/COperatorRGB.h"

#include <cmath>

//using namespace std;
using namespace TUVision;

void
TUVision::calcOpticalFlowUnnormalized(const CImageByteGray& orig, 
                                      const CImageByteGray& next, 
                                      CImageIntGray** o_fx, 
                                      CImageIntGray** o_fy)  
{
    CImageByteGray win1(orig,false);
    CImageByteGray win2(orig,false);

    win1.setWindow(0,0,win1.getW()-1, win1.getH());
    win2.setWindow(1,0,win2.getW()-1, win2.getH());

    CImageIntGray diffX(win1-win2);

    win1.resetWindow();
    win2.resetWindow();

    win1.setWindow(0,0,win1.getW(), win1.getH()-1);
    win2.setWindow(0,1,win2.getW(), win2.getH()-1);

    CImageIntGray diffY(win1-win2);

    CImageIntGray diffT(orig-next);


    int w;
    int h;

    w=orig.getW()-1;
    h=orig.getH()-1;

    *o_fx = new CImageIntGray (w,h);
    *o_fy = new CImageIntGray (w,h);


    int i,j;
    for(j=0;j<h;++j) {
        const int* dx=diffX.getRow(j);
        const int* dy=diffY.getRow(j);
        const int* dt=diffT.getRow(j);

        int *ofx=(*o_fx)->getRow(j);
        int *ofy=(*o_fy)->getRow(j);


        // -dI/dt * \nabla I (without normalizing: (1.0/ (\norm { \nabla I }^2)))
        for(i=0;i<w;++i) {
            *ofx++ = -*dt   * *dx++;
            *ofy++ = -*dt++ * *dy++;
        }

    }
}

void
TUVision::calcOpticalFlow(const CImageByteGray& orig, 
                          const CImageByteGray& next, 
                          CImageDoubleGray** o_fx, 
                          CImageDoubleGray** o_fy)  
{
    CImageByteGray win1(orig,false);
    CImageByteGray win2(orig,false);

    win1.setWindow(0,0,win1.getW()-1, win1.getH());
    win2.setWindow(1,0,win2.getW()-1, win2.getH());

    CImageIntGray diffX(win1-win2);

    win1.resetWindow();
    win2.resetWindow();

    win1.setWindow(0,0,win1.getW(), win1.getH()-1);
    win2.setWindow(0,1,win2.getW(), win2.getH()-1);

    CImageIntGray diffY(win1-win2);

    CImageIntGray diffT(orig-next);


    int w;
    int h;

    w=orig.getW()-1;
    h=orig.getH()-1;

    *o_fx = new CImageDoubleGray (w,h);
    *o_fy = new CImageDoubleGray (w,h);


    int i,j;
    for(j=0;j<h;++j) {
        const int* dx=diffX.getRow(j);
        const int* dy=diffY.getRow(j);
        const int* dt=diffT.getRow(j);

        double *ofx=(*o_fx)->getRow(j);
        double *ofy=(*o_fy)->getRow(j);

        int x, y;
        double len;

        // -dI/dt * \nabla I / (\norm { \nabla I }^2)
        for(i=0;i<w;++i) {
            x=*dx++;
            y=*dy++;
            len=sqrt((double)x*x + y*y);

            *ofx++ = (double)-*dt   * x/len;
            *ofy++ = (double)-*dt++ * y/len;
        }

    }
}


CImageByteRGB 
TUVision::beautifyOpticalFlow(const CImageIntGray& o_fx, 
                              const CImageIntGray& o_fy)
{
    int w;
    int h;

    w=o_fx.getW();
    h=o_fx.getH();

    if(w != o_fy.getW() || h != o_fy.getH()) {

        // i know, i know.
        return(CImageByteRGB(1,1));
    }

    CImageByteRGB rgb(o_fx.getW(), o_fx.getH());

    int i,j;
    for(j=0;j<h;++j) {
        const int* ofx=o_fx.getRow(j);
        const int* ofy=o_fy.getRow(j);

        unsigned char* dst=rgb.getRow(j);

        for(i=0;i<w;++i) {
            int x=*ofx++;
            int y=*ofy++;

            if(x==0 && y==0) {
                *dst++=128;
                *dst++=0;
                *dst++=128;
                continue;
            }
            double len=(double)x*x + y*y;


            *dst++=(unsigned char)(127.0*x/len+128);
            *dst++=0;
            *dst++=(unsigned char)(127.0*y/len+128);
        }

    }
    return(CImageByteRGB(rgb,false));
}
