/***************************************************************************
                          CCameraMatrix.cpp  -  description
                             -------------------
    begin                : 12.03.2003
    copyright            : (C) 2003 by Thomas Beilicke
    email                : tuvisionlib@tuvision.de
    origin               : Diplomarbeit Thomas Beilicke
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#include "tuv3d/CCameraMatrix.h"
#include "newmat/newmatap.h"

using namespace TUVision;

CVertex3d CCameraMatrix::upVector(0, 1, 0);

/**********************************************************/ 
/*! The camera is located at the world coordinate system.   
 */
CCameraMatrix::CCameraMatrix() : NEWMAT::Matrix(3,4)
{
    // Create canonical camera matrix
    (*this)	<< 1 << 0 << 0 << 0
        << 0 << 1 << 0 << 0
        << 0 << 0 << 1 << 0;
}

/**********************************************************/ 
/*! The parameters are explained for createCamera().
 */
CCameraMatrix::CCameraMatrix(const CPoint3D<double> &origin, const CPoint3D<double> &lookat, 
                             double aspect, double focal, double kx, double u0, double v0, double skew) 
                             : NEWMAT::Matrix(3,4)
{
    createCamera(origin, lookat, aspect, focal, kx, u0, v0, skew);
}

/**********************************************************/ 
/*! The NEWMAT matrix should contain an actual camera matrix.
 *
 *	\throws NEWMAT::IncompatibleDimensionException if the matrix
 *			is not 3x4.
 *
 *	\param P The NEWMAT matrix to copy.
 */
CCameraMatrix::CCameraMatrix(const NEWMAT::Matrix &P) : Matrix(3,4)
{
    // if P has wrong size, SubMatrix throws an IncompatibleDimensionsException
    SubMatrix(1,3,1,4)=P;
}

/**********************************************************/ 
/*! The NEWMAT matrix should contain an actual camera matrix.
 *
 *	\throws NEWMAT::IncompatibleDimensionException if the matrix
 *			is not 3x4.
 *
 *	\param P The NEWMAT matrix to copy.
 */
CCameraMatrix &CCameraMatrix::operator=(const NEWMAT::Matrix &P)
{
    // if P has wrong size, SubMatrix throws IncompatibleDimensionsException
    SubMatrix(1,3,1,4)=P;
    return (*this);
}

/**********************************************************/ 
/*! The parameters and functionality are explained for createCamera().
 */
CCameraMatrix::CCameraMatrix(const std::vector<CCorrespondence> &corresp, 
                             unsigned int index) : NEWMAT::Matrix(3,4)
{

    calculateCamera(corresp,index);
}

/**********************************************************/ 
/*! This algorithm uses the euclidean coordinates of 6
 *  world points and the corresponding projections.
 *  The according camera matrix is calculated out of 
 *  this information. 
 *
 *  \remark Ransac and Minpack should be 
 *    used additionally if noise is present! It is not checked
 *    whether the correspondence contains the 6 points needed!
 *
 *	\param corresp A vector of correspondences.
 *         The world point and the image point #index are
 *         used. Only 6 points are needed but all are used.
 *  \param index The correspondence's image point to use.
 */
void CCameraMatrix::calculateCamera(const std::vector<CCorrespondence> &corresp, unsigned int index)
{

    // Need 6 point matches, but using all available
    NEWMAT::Matrix A(corresp.size()*2,12);    
    unsigned int i;
    std::vector<CCorrespondence>::const_iterator iter;
    for (i=0, iter=corresp.begin(); i<corresp.size(); ++iter, ++i) {
        CVertex3d wpt=iter->getEuclideanWorldPoint();
        CVertex2d ipt=iter->getEuclideanImagePoint(index);
        // Equations are based on image and world points in euclidean coordinates
        A.Row(2*i+1)    << 0 << 0 << 0 << 0 << -wpt.x << -wpt.y << -wpt.z << -1.0 
            << wpt.x * ipt.y << wpt.y * ipt.y << wpt.z * ipt.y << 1.0 * ipt.y;
        A.Row(2*(i+1))  << wpt.x << wpt.y << wpt.z << 1.0 << 0 << 0 << 0 << 0
            << -wpt.x * ipt.x << -wpt.y * ipt.x << -wpt.z * ipt.x << -1.0 * ipt.x;
    }

    NEWMAT::DiagonalMatrix D; NEWMAT::Matrix U,V;
    NEWMAT::SVD(A,D,U,V);
    NEWMAT::ColumnVector solve=(NEWMAT::ColumnVector) V.Column(12);

    (*this) << solve(1) << solve(2) << solve(3) << solve(4)
        << solve(5) << solve(6) << solve(7) << solve(8)
        << solve(9) << solve(10) << solve(11) << solve(12);    

    // Cam is determined up to scale => might be multiplied by -1 !
    // scale ambiguity can be removed because of intrinsic's format
    CIntrinsicMatrix K_; NEWMAT::Matrix R(3,3);
    getIntrinsicRotation(K_, R);
    double scale=1.0/K_(3,3);  // Remove scale ambiguity, intrinsic(3,3) must be 1.0

    if (R.Determinant() < 0.0)
        scale*=-1.0;    // Change sign, if rotation matrix also mirrors!

    (*this)*=scale;   // Scale this matrix
}

/**********************************************************/ 
/*! This is used as quality factor for CRansac.
 *  The deviation between the projected world point
 *  and the corresponding image point is calculated (geometric
 *  distance).
 *
 *  \remark CRansac calls this function directly.
 *
 *	\param sample A correspondence.
 *         The world point and the image point #index are
 *         used. 
 *  \param index The correspondence's image point to use.
 */
double CCameraMatrix::calculateDistance( const TDataVector::value_type &sample, unsigned int index)
{
    CVertex4d wpt=sample.getWorldPoint();
    CVertex2d ipt=sample.getEuclideanImagePoint(index);
    CVertex3d ppt=project(wpt);
    ppt/=ppt.z;    // To euclidean

    //    return (ipt.x-ppt.x)*(ipt.x-ppt.x)+(ipt.y-ppt.y)*(ipt.y-ppt.y); // projection error, geometric distance
    double tmp1=ipt.x-ppt.x;
    double tmp2=ipt.y-ppt.y;
    return sqrt(tmp1*tmp1+tmp2*tmp2); // Same as above, but faster (and now using sqrt)
}

/**********************************************************/ 
/*! The previous camera matrix is overwritten. The camera
 *  orientation is given by the world-up-vector.
 *  If it cannot be calculated, the z-axis is used instead
 *  afterwards the x-axis is tried.
 *
 *	\param origin The camera centre position in euclidean 
 *				  world coordinates.
 *	\param lookat The target of the camera in euclidean
 *				  world coordinates.
 *	\param aspect The aspect ration: aspect=ky/kx
 *	\param focal  The focal length
 *	\param kx	  Pixels per unit distance in x.
 *	\param u0	  The optical axis in the image (x, in pixels)
 *	\param v0	  The optical axis in the image (y, in pixels)
 *	\param skew   The skew of the camera.
 */
void CCameraMatrix::createCamera(const CPoint3D<double> &origin, const CPoint3D<double> &lookat, 
                                 double aspect, double focal, double kx, double u0, double v0, double skew)
{

    CPoint3D<double> d=lookat-origin; // d shows from origin to camera to calculate the angles right

    NEWMAT::ColumnVector world_up=upVector.toColumnVector();

    // 3rd matrix row
    NEWMAT::ColumnVector out=d.toColumnVector();
    out/=out.NormFrobenius();

    // 2nd matrix row
    NEWMAT::ColumnVector up=world_up - (world_up.t()*out).AsScalar()*out;
    if (up.NormFrobenius() < 1e-5) {  // Too small => parallel to world_up        
        world_up << 0 << 0 << 1;   // Use world z-axis instead
        up=world_up - (world_up.t()*out).AsScalar()*out;
    }
    if (up.NormFrobenius() < 1e-5) {  // This can happen if the world_up is set to z-axis
        world_up << 1 << 0 << 0;   // Use world x-axis instead
        up=world_up - (world_up.t()*out).AsScalar()*out;
    }   
    up/=up.NormFrobenius();

    // 1st matrix row
    NEWMAT::ColumnVector right(3);
    right(1)=up(2)*out(3)-up(3)*out(2);
    right(2)=-up(1)*out(3)+up(3)*out(1);
    right(3)=up(1)*out(2)-up(2)*out(1);
    right/=right.NormFrobenius();

    // Rotation matrix 
    NEWMAT::Matrix R(3,3);
    R.Row(1) = right.t();
    R.Row(2) = up.t();
    R.Row(3) = out.t();

    NEWMAT::ColumnVector t=-R*origin.toColumnVector();
    NEWMAT::Matrix A(3,3);
    A
        << kx*focal << skew << u0
        << 0 << kx*aspect*focal << v0
        << 0 << 0 << 1;
    NEWMAT::Matrix RT(3,4);
    RT.SubMatrix(1,3,1,3)=R;
    RT.Column(4)=t;

    SubMatrix(1,3,1,4)=A*RT;	
}

/**********************************************************/ 
/*! The camera up-vector and the camera lookat-vector 
 *  are coplanar with this world-up-vector.
 * 
 *	\param upvector The up vector of the world coordinate system.
 */
void CCameraMatrix::setUpVector(const CPoint3D<double> &upvector)
{
    upVector=upvector;
}

/**********************************************************/ 
/*! This function is based on Hartley/Zisserman: 
 *  Multiple View Geometry in Computer Vision, p.145
 *
 *	\warning This function only works for finite cameras! 
 *
 *	\returns A four dimensional (homogeneous) vector.
 */
NEWMAT::ColumnVector CCameraMatrix::getCentre() const
{
    // Decompose camera matrix
    NEWMAT::Matrix M=SubMatrix(1,3,1,3).i();
    NEWMAT::ColumnVector p=Column(4);

    NEWMAT::ColumnVector result(4);
    result.SubMatrix(1,3,1,1)=-M*p;	// Create homogeneous coordinates
    result(4)=1.0;

    return result;
}

/**********************************************************/ 
/*! 
 *	\returns The projected three dimensional (homogeneous) vector.
 */
NEWMAT::ColumnVector CCameraMatrix::project(const NEWMAT::ColumnVector &point) const
{
    return operator*(point);
}

/**********************************************************/ 
/*! 
 *	\returns The projected three dimensional (homogeneous) vector.
 */
NEWMAT::ColumnVector CCameraMatrix::project(const CPoint4D<double> &point) const 
{
    return operator*(point.toColumnVector());
}

/**********************************************************/ 
/*! Both matrices are calculated at the same time. So if both
 *  matrices are needed, this function should be called.
 *  This function is based on Hartley/Zisserman p.150
 *
 *	\param K An intrinsic matrix. It is overwritten by this
 *			  camera's intrinsic 3x3 matrix.
 *	\param R A NEWMAT matrix. It is overwritten by this
 *			  camera's 3x3 orthogonal rotation matrix.
 *
 *	\remark Note that P != K*R. Use getIntrinsicExtrinsic instead.
 */
void CCameraMatrix::getIntrinsicRotation(CIntrinsicMatrix &K, NEWMAT::Matrix &R) const 
{
    rq(K,R);

    // remove ambiguity/ positive diagonal entries for K
    if (K(1,1)<0) {
        K.Column(1)*=-1.0;
        R.Row(1)*=-1.0;
    }

    if (K(2,2)<0) {
        K.Column(2)*=-1.0;
        R.Row(2)*=-1.0;
    }

    if (K(3,3)<0) {	// should be 1.0
        K.Column(3)*=-1.0;
        R.Row(3)*=-1.0;
    }

}


/**********************************************************/ 
/*! Both matrices are calculated at the same time. So if both
 *  matrices are needed, this function should be called.
 *  This function calls getIntrinsicRotation() and getCentre()
 *  to build the extrinsic matrix such that P=K*E. E is 
 *  calculated according to Hartley/Zisserman p.150
 *
 *	\param K An intrinsic matrix. It is overwritten by this
 *			  camera's intrinsic 3x3 matrix.
 *	\param E A NEWMAT matrix. It is overwritten by this
 *			  camera's 3x4 extrinsic matrix. 
 */
void CCameraMatrix::getIntrinsicExtrinsic(CIntrinsicMatrix  &K, NEWMAT::Matrix &E) const
{
    NEWMAT::Matrix tmp;
    getIntrinsicRotation(K,tmp);
    E=NEWMAT::Matrix(3,4);
    E.SubMatrix(1,3,1,3)=tmp;
    E.Column(4)=-tmp*getCentre().Rows(1,3);
}

/**********************************************************/ 
/*! Note that the intrinsic matrix is calculated by this
 *  decomposition. So if you need both use 
 *  getIntrinsicExtrinsic() instead!
 *
 *	\param E A NEWMAT matrix. It is overwritten by this
 *			  camera's 3x4 extrinsic matrix. 
 */
void CCameraMatrix::getExtrinsic(NEWMAT::Matrix &E) const 
{
    CIntrinsicMatrix K;
    getIntrinsicExtrinsic(K, E);
}

/**********************************************************/ 
/*! Note that the rotation matrix R is calculated by this
 *  decomposition. So if you need both use 
 *  getIntrinsicRotation() instead!
 *
 *	\param K An intrinsic matrix. It is overwritten by this
 *			  camera's intrinsic 3x3 matrix.
 */
void CCameraMatrix::getIntrinsic(CIntrinsicMatrix &K) const 
{
    NEWMAT::Matrix R;
    getIntrinsicRotation(K, R);
}

/**********************************************************/ 
/*! Note that the intrinsic matrix is calculated by this
 *  decomposition. So if you need both use 
 *  getIntrinsicExtrinsic() instead!
 *
 *	\param R A NEWMAT matrix. It is overwritten by this
 *			  camera's 3x3 orthogonal rotation matrix.
 */
void CCameraMatrix::getRotation(NEWMAT::Matrix &R) const 
{
    CIntrinsicMatrix K;
    getIntrinsicRotation(K, R);
}

/**********************************************************/ 
/*! The principal axis is the axis from the camera centre
 *  perpendicular to the image plane. It points away from
 *  the camera centre (it represents the direction the
 *  camera is looking at).
 *  Multiple View Geometry in Computer Vision, p.148
 *
 *  \remark The returned vector is not normalized!
 *
 *	\warning This function only works for finite cameras! 
 *
 *	\returns A three dimensional vector in the world coordinate 
 *    system.
 */
NEWMAT::ColumnVector CCameraMatrix::getPrincipalAxis() const 
{
    NEWMAT::Matrix M=SubMatrix(1,3,1,3);
    NEWMAT::ColumnVector m3=M.Row(3).t();
    return M.Determinant()*m3;
}

/**********************************************************/ 
/*! The principal point is the point in the image plane
 *  where the principal axis passes through.
 *  Based on Multiple View Geometry in Computer Vision, p.147
 *
 *  \remark This coordinates can also be retrieved by
 *	 evaluating the intrinsic parameters u0 and v0.
 *
 *	\returns A three dimensional (homogeneous) vector in the
 *   camera coordinate system.
 */
NEWMAT::ColumnVector CCameraMatrix::getPrincipalPoint() const 
{
    return SubMatrix(1,3,1,3)*SubMatrix(3,3,1,3).t();
}

/**********************************************************/ 
/*! Note that: P*worldpoint=projectedpoint must be fulfilled
 *  directly. So if depth is positive, worldpoint lies in front of 
 *  the camera. This is the distance from the camera
 *  centre to the world point. The calculation is based
 *  on Hartley/Zisserman p.149. Note the other depth function available 
 *  (This method has slightly less calulative effort)!
 *
 *  \warning If you have calculated the world point X by a
 *			 triangulation the projected point x has to be
 *			 recalculated as x=P*X to restore its direct dependance
 *           on X!!
 * 
 *  \remark This calculation is independent of the homogeneous 
 *   representation of the worldpoint and the camera matrix.
 *   After the projection the projected point may not be normalized 
 *   before the depth is calculated! 
 *
 *  \param worldpoint The worldpoint X (that is projected to x) 
 *			in homogeneous coordinates (4D).
 *
 *	\returns The depth of the worldpoint X.
 */
double CCameraMatrix::depth(const NEWMAT::ColumnVector &worldpoint, const NEWMAT::ColumnVector &projectedpoint) const 
{
    NEWMAT::Matrix M=SubMatrix(1,3,1,3);
    double depth=projectedpoint(3)/worldpoint(4)/M.Row(3).NormFrobenius();
    if (M.Determinant()>0)
        return depth;
    else
        return -depth;
}

/**********************************************************/ 
/*! Note that: Contrary to the other depth function (which
 *  is slightly faster) no prerequisite has to be considered 
 *  here. Note that the depths can only be compared if X
 *  is a euclidean point or was created by euclidean 
 *  reconstruction. 
 *
 *  \remark This calculation is independent of the homogeneous 
 *   representation of the worldpoint and the camera matrix.
 *
 *  \param worldpoint The worldpoint X in homogeneous 
 *         coordinates (4D).
 *
 *	\returns The depth of the worldpoint X.
 */
double CCameraMatrix::depth(const NEWMAT::ColumnVector &worldpoint) const 
{
    NEWMAT::ColumnVector m=getPrincipalAxis();
    m/=m.NormFrobenius();
    NEWMAT::ColumnVector X=worldpoint;
    X/=X(4);	// To euclidean

    NEWMAT::ColumnVector C=getCentre();
    X-=C;
    double depth=(m.t()*X.Rows(1,3)).AsScalar();
    return depth;
}

/**********************************************************/ 
/*! This function calculates K,R such that P=K*R. The 
 *  algorithm uses Givens Rotations, as shown in 
 *  Hartley/Zisserman p.552.
 *
 *	\param K A NEWMAT matrix. It is overwritten by an upper triangular matrix.
 *	\param E A NEWMAT matrix. It is overwritten by an orthogonal matrix.
 *
 *  \remark This implementation only works for 3x3 matrices! The diagonal
 *			entries of K are not necissarily positive!
 */
void CCameraMatrix::rq(NEWMAT::Matrix &K, NEWMAT::Matrix &R) const 
{
    R=NEWMAT::IdentityMatrix(3);
    K=SubMatrix(1,3,1,3);	// Only operate on 3x3 SubMatrix of P
    NEWMAT::Matrix Q(3,3);

    //set P32 to zero:
    double a=K(3,2);
    double b=K(3,3);
    double c;
    double s;

    if (a!=0) {
        c=-b/sqrt(a*a+b*b);
        s=a/sqrt(a*a+b*b);
        Q	// Givens rotation
            <<  1 <<  0 <<  0
            <<  0 <<  c << -s
            <<  0 <<  s <<  c;
        R=Q.t()*R;
        K*=Q;
    }

    // set P31 to zero:
    a=K(3,1);
    b=K(3,3);

    if (a!=0) {
        c=b/sqrt(a*a+b*b);
        s=a/sqrt(a*a+b*b);
        Q	// Givens rotation
            <<  c <<  0 <<  s
            <<  0 <<  1 <<  0
            << -s <<  0 <<  c;
        R=Q.t()*R;
        K*=Q;
    }

    // set P21 to zero:
    a=K(2,1);
    b=K(2,2);

    if (a!=0) {
        c=-b/sqrt(a*a+b*b);
        s=a/sqrt(a*a+b*b);
        Q	// Givens rotation
            <<  c << -s <<  0
            <<  s <<  c <<  0
            <<  0 <<  0 <<  1;
        R=Q.t()*R;
        K*=Q;
    }
}
