/***************************************************************************
CEssentialMatrix.cpp  -  description
-------------------
begin                : 02.04.2003
copyright            : (C) 2003 by Thomas Beilicke
email                : tuvisionlib@tuvision.de
origin               : Diplomarbeit Thomas Beilicke
***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#include "tuv3d/CEssentialMatrix.h"
#include "tuv3d/CTriangulation.h"
#include <vector>

using namespace TUVision;

/**********************************************************/ 
/*! The matrix does not really make sense, but the matrix
 *  should be filled when using the default constructor
 *  and it is an essential matrix...
 */
CEssentialMatrix::CEssentialMatrix() : NEWMAT::Matrix(3,3)
{
    // Create somehow canonical matrix, does not really make sense...
    (*this)	<< 1 << 0 << 0
        << 0 << 1 << 0
        << 0 << 0 << 0;
}

/**********************************************************/ 
/*! The parameters are explained for calculateMatrix().
 */
CEssentialMatrix::CEssentialMatrix(const CFundamentalMatrix &F, const CIntrinsicMatrix &K1, 
                                   const CIntrinsicMatrix &K2)	: NEWMAT::Matrix(3,3) 
{
    calculateMatrix(F, K1, K2);
}

/**********************************************************/ 
/*! The parameters are explained for calculateMatrix().
 */
CEssentialMatrix::CEssentialMatrix(const CFundamentalMatrix &F, const CCameraMatrix &P1, 
                                   const CCameraMatrix &P2) : NEWMAT::Matrix(3,3) 
{
    calculateMatrix(F, P1, P2);
}

/**********************************************************/ 
/*! The NEWMAT matrix should contain an actual essential matrix.
 *
 *	\throws NEWMAT::IncompatibleDimensionException if the matrix
 *			is not 3x3.
 *
 *	\param E The NEWMAT matrix to copy.
 */
CEssentialMatrix::CEssentialMatrix(const NEWMAT::Matrix &E) : Matrix(3,3) 
{
    // if E has wrong size, SubMatrix throws an IncompatibleDimensionsException
    SubMatrix(1,3,1,3)=E;
}

/**********************************************************/ 
/*! The NEWMAT matrix should contain an actual essential matrix.
 *
 *	\throws NEWMAT::IncompatibleDimensionException if the matrix
 *			is not 3x3.
 *
 *	\param E The NEWMAT matrix to copy.
 */
CEssentialMatrix &CEssentialMatrix::operator=(const NEWMAT::Matrix &E) 
{
    // if E has wrong size, SubMatrix throws IncompatibleDimensionsException
    SubMatrix(1,3,1,3)=E;
    return (*this);
}

/**********************************************************/ 
/*! The previous essential matrix is overwritten. The
 *  fundamental matrix for two images is needed as well
 *  as both cameras' intrinsic parameters.
 *
 *	\param F The fundemental matrix.
 *	\param K1 The intrinsic parameter matrix of the first 
 *			   camera.
 *	\param K2 The intrinsic parameter matrix of the second
 *			   camera. 
 */
void CEssentialMatrix::calculateMatrix(const CFundamentalMatrix &F, const CIntrinsicMatrix &K1, 
                                       const CIntrinsicMatrix &K2)
{

    SubMatrix(1,3,1,3)=K2.t()*F*K1;	
}

/**********************************************************/ 
/*! The previous essential matrix is overwritten. The
 *  fundamental matrix for two images is needed as well
 *  as both camera matrices. This function calculates
 *  the intrinsic matrices and calls calculateMatrix().
 *
 *  \remark This function only makes sense if the cameras
 *			are calibrated. 
 *
 *	\param F The fundemental matrix.
 *	\param P1 The first camera matrix.
 *	\param P2 The second camera matrix.
 */
void CEssentialMatrix::calculateMatrix(const CFundamentalMatrix &F, const CCameraMatrix &P1, 
                                       const CCameraMatrix &P2)
{
    CIntrinsicMatrix K1, K2;
    P1.getIntrinsic(K1);
    P2.getIntrinsic(K2);
    calculateMatrix(F, K1, K2);	
}

/**********************************************************/ 
/*! The calculated matrix may not be an essential matrix
 *  due to numerical problems. This function calculates
 *  a matrix E' such that |E'-E| is minimized (Frobeniusnorm).
 *  Based on Faugeras: Geometry of Multiple Images, p.284
 *
 *  \remark The essential matrix is replaced by the calculated one.
 *
 */
void CEssentialMatrix::closestEssential() 
{
    // Decompose essential:
    NEWMAT::Matrix U,V;
    NEWMAT::DiagonalMatrix D;
    NEWMAT::SVD(*this,D,U,V);
    double d=(D(1,1)+D(2,2))/2.0;
    D(1,1)=d;
    D(2,2)=d;
    D(3,3)=0;
    (*this)=U*D*V.t();
}

/**********************************************************/ 
/*! The relative position between two cameras can be calculated
 *  out of the essential matrix. This function returns the second
 *  camera. The first camera is the canonical camera multiplied
 *  by the appropriate intrinsic matrix. This algorithm needs
 *  one normalized point correspondence to project a test point.
 *  Based on Hartley/Zisserman, p. 239
 *
 *  \remark Calling closesEssential() to improve the essential
 *			matrix has no effect on this function.
 *
 *  \param K The intrinsic matrix belonging to the 2nd camera.
 *	\param normalizedcorresp A normalized correspondence of one
 *		   point from the correspondence data. (x_norm=inv(K)*x).
 *	\param camNo1 The number of the first camera. A correspondence can 
 *			contain the image points of many cameras. The algorithm 
 *			will use the correspondences number camNo1 and camNo2 for 
 *			triangulation. The camera number starts with 0 as in 
 *			CWorldCorrespondence. The default is the camerapair 0 and 1.
 *	\param camNo2 Same as above for the 2nd camera.
 * 
 *  \returns The 2nd camera found. It is already multiplied with
 *			 its intrinsic matrix.
 */
CCameraMatrix CEssentialMatrix::calculateCamera(const TUVision::CIntrinsicMatrix &K, 
                                                const TUVision::CCorrespondence &normalizedcorresp, 
                                                unsigned int camNo1, unsigned int camNo2) const
{
    // Decompose essential matrix:
    NEWMAT::Matrix U,V;
    NEWMAT::DiagonalMatrix D;
    NEWMAT::SVD( (*this),D,U,V);

    // E is determined up to scale. If det(U)!=det(V), 
    // then U*W*V' yields a rotation+mirror
    // To prevent this: Setting U=-U 
    // => E=-E which is OK, because E is determined up to scale!
    if (U.Determinant()*V.Determinant() < 0) 
        U=-U;

    NEWMAT::Matrix W(3,3);
    W
        << 0 << -1 << 0
        << 1 <<  0 << 0
        << 0 <<  0 << 1;

    NEWMAT::Matrix R=U*W*V.t();			// Rotation
    NEWMAT::Matrix R_=U*W.t()*V.t();	// Rotation
    NEWMAT::ColumnVector t=U.Column(3);	// Translation

    // 4 Possible 2nd cameras:
    std::vector<CCameraMatrix> trycameras;	
    CCameraMatrix C;

    C.SubMatrix(1,3,1,3)=R;
    C.SubMatrix(1,3,4,4)=t;
    trycameras.push_back(C);	// C=[R  | t]

    C.SubMatrix(1,3,4,4)=-t;
    trycameras.push_back(C);	// C=[R  |-t]

    C.SubMatrix(1,3,1,3)=R_;
    trycameras.push_back(C);	// C=[R_ |-t]

    C.SubMatrix(1,3,4,4)=t;
    trycameras.push_back(C);	// C=[R_ | t]

    // Calculate cameras
    int result=0;
    CCorrespondence corresp=normalizedcorresp;	// Copy normalized correspondence to a tmp-worldcorrespondence

    C=CCameraMatrix(); // canonical camera matrix (=> Camera 1)
    double d1, d2;
    int i;

    std::vector<CCameraMatrix>::iterator camIter;
    for (camIter=trycameras.begin(), i=1; camIter<trycameras.end(); ++camIter, ++i) {	
        // Reproject normalized "test" point
        // polynomial ... to be used later... ??
        //CTriangulation::polynomial(C, *camIter, (*this), corresp, 0, 1))
        CTriangulation::iterativeEigen(C, *camIter, corresp, 15, 0, 1);

        // Determine whether test point is in front of both cameras
        d1=C.depth(corresp.getWorldPoint().toColumnVector());
        d2=(*camIter).depth(corresp.getWorldPoint().toColumnVector());
        if (d1>0 && d2>0 && result!=0)	// More than one valid solution found
            throw CExceptionEssential("More than one possible 2nd camera found!", CExceptionEssential::TOOMANYSOLUTIONS);
        else if (d1>0 && d2>0)
            result=i;
    }

    if (result==0)	// No valid solution found
        throw CExceptionEssential("No possible camera found!", CExceptionEssential::NOSOLUTION);

    return TUVision::CCameraMatrix(K*trycameras[result-1]);
}
