/***************************************************************************
                          cfundamentalmatrix.cpp  -  description
                             -------------------
    begin                : Die Apr 9 2002
    copyright            : (C) 2002 by Fabian Wenzel
    email                : wenzel@tu-harburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 * This program is part of Fabian Wenzel's PhD - project. It may not be    *
 * used otherwise. It may not be used otherwise.                           *
 *                                                                         *
 ***************************************************************************/

#include "tuv3d/CFundamentalmatrix.h"
#include "newmat/newmatap.h"
#include <complex>
#include <algorithm>

using namespace NEWMAT;
using namespace std;
using namespace TUVision;

/***************************************************************************/
/*!
 *  Initializes the matrix-part by 3x3 and calculates its values based
 *  on the number of input data items
 *
 ***************************************************************************/
CFundamentalMatrix::CFundamentalMatrix( const CFundamentalMatrix::TDataVector &list,
                                       unsigned int firstindex, unsigned int secondindex) :
Matrix(3,3)
{
    if (list.size() == 7)
        sevenPointAlgorithm(list, firstindex, secondindex);

    else if (list.size() > 7)
        eightPointAlgorithm(list, firstindex, secondindex);

}

/**********************************************************/ 
/*! The matrix contents do not really make sense, but 
 *  it is a fundamental matrix. Useful for creating
 *  a fundamental matrix that is filled later.
 */
CFundamentalMatrix::CFundamentalMatrix() : NEWMAT::Matrix(3,3) {
    // Create somehow canonical matrix, does not really make sense...
    (*this) << 1 << 0 << 0
        << 0 << 1 << 0
        << 0 << 0 << 0;
}

/**********************************************************/ 
/*! The NEWMAT matrix should contain an actual fundamental matrix.
 *
 *	\throws NEWMAT::IncompatibleDimensionException if the matrix
 *			is not 3x3.
 *
 *	\param F The NEWMAT matrix to copy.
 */
CFundamentalMatrix::CFundamentalMatrix(const NEWMAT::Matrix &F) : Matrix(3,3) {
    // if F has wrong size, SubMatrix throws an IncompatibleDimensionsException
    SubMatrix(1,3,1,3)=F;
}

/**********************************************************/ 
/*! The NEWMAT matrix should contain an actual fundamental matrix.
 *
 *	\throws NEWMAT::IncompatibleDimensionException if the matrix
 *			is not 3x3.
 *
 *	\param F The NEWMAT matrix to copy.
 */
CFundamentalMatrix &CFundamentalMatrix::operator=(const NEWMAT::Matrix &F) {
    // if F has wrong size, SubMatrix throws IncompatibleDimensionsException
    SubMatrix(1,3,1,3)=F;
    return (*this);
}

/***************************************************************************/
/*!
 *  Destructor does nothing as there is no dynamic data
 *
 ***************************************************************************/
CFundamentalMatrix::~CFundamentalMatrix()
{
}

/**********************************************************/ 
/*! The constructor just calls the create() function. See 
 *  documentation below.
 **********************************************************/ 
CFundamentalMatrix::CFundamentalMatrix (const CCameraMatrix &P1, const CCameraMatrix &P2) 
: Matrix(3,3){
    create(P1, P2);
}

/**********************************************************/ 
/*! This function should only be used for synthetical 
 * data and testing purposes. Based on Hartley/Zisserman p226.
 *
 *
 *	\param P1 The first camera matrix.
 *      \param P2 The second camera matrix.
 **********************************************************/ 
void CFundamentalMatrix::create(const CCameraMatrix &P1, const CCameraMatrix &P2) {
    ColumnVector e2=P2*P1.getCentre();	// Epipole from c1 in camera 2
    Matrix E(3,3);
    E							// Rebuild cross product
        << 0 << -e2(3) << e2(2)
        << e2(3) << 0 << -e2(1)
        << -e2(2) << e2(1) << 0;    

    Matrix PseudoInvP1;
    Matrix U,V;
    DiagonalMatrix D;

    // Build Pseudoinverse
    SVD(P1.t(), D, U, V);   // NEWMAT-SVD only works for (m,n) matrices with m>n
    for (int i=1; i<=D.Nrows(); ++i)
        if (D(i)!=0)
            D(i)=1.0/D(i);
    PseudoInvP1=(V*D*U.t()).t();

    SubMatrix(1,3,1,3)=E*P2*PseudoInvP1;
}

/***************************************************************************/
/*!
 *  Computer its internal values based on the seven-point algorithms.
 *  The results will not be absolutely good but is well suited for
 *  e.g. RANSAC. (see Pollefeys-Tutorial, page 39)
 *
 ***************************************************************************/
void CFundamentalMatrix::sevenPointAlgorithm(const CFundamentalMatrix::TDataVector &list,
                                             unsigned int firstindex, unsigned int secondindex)
{
    CFundamentalMatrix::TDataVector::const_iterator currentpointpair;

    Matrix A(9,9);
    int i;
    for(i=1, currentpointpair  = list.begin();
        currentpointpair != list.end()  ;
        ++currentpointpair,++i)
    {
        A.Row(i) << (*currentpointpair)[firstindex].x * (*currentpointpair)[secondindex].x
            << (*currentpointpair)[firstindex].y * (*currentpointpair)[secondindex].x
            <<                            (*currentpointpair)[secondindex].x
            << (*currentpointpair)[firstindex].x * (*currentpointpair)[secondindex].y
            << (*currentpointpair)[firstindex].y * (*currentpointpair)[secondindex].y
            <<                            (*currentpointpair)[secondindex].y
            << (*currentpointpair)[firstindex].x
            << (*currentpointpair)[firstindex].y
            << 1;
    }

    // Add zeros such that A is square. Reason: NEWMAT's SVD only works with
    // #rows >= #columns. Using A' is not possible as the dimension of U,D,V'
    // is different from e.g. MATLAB's implementation, so backconversion is not
    // possible easily

    A.Row(8) << 0 << 0 << 0 << 0 << 0 << 0 << 0 << 0 << 0;
    A.Row(9) << 0 << 0 << 0 << 0 << 0 << 0 << 0 << 0 << 0;

    Matrix         V;
    Matrix         U;
    DiagonalMatrix D;

    SVD(A,D,U,V);

    // Transpose matrix as row access will be faster (see below)
    V = V.t();

    // Calculate Coefficients for root-finding (see Pollefeys-Tutorial, page 37- (4.8) )
    // This is very annoying as there are many coefficients

    double a3 = V(9,1)*(V(9,5)*V(9,9)-V(9,8)*V(9,6))
        -V(9,2)*(V(9,4)*V(9,9)-V(9,7)*V(9,6))
        +V(9,3)*(V(9,4)*V(9,8)-V(9,7)*V(9,5));

    double a2 = V(8,1)*(V(9,5)*V(9,9)-V(9,6)*V(9,8))
        -V(8,2)*(V(9,4)*V(9,9)-V(9,7)*V(9,6))
        +V(8,3)*(V(9,4)*V(9,8)-V(9,5)*V(9,7))
        -V(8,4)*(V(9,2)*V(9,9)-V(9,3)*V(9,8))
        +V(8,5)*(V(9,1)*V(9,9)-V(9,3)*V(9,7))
        -V(8,6)*(V(9,1)*V(9,8)-V(9,2)*V(9,7))
        +V(8,7)*(V(9,2)*V(9,6)-V(9,3)*V(9,5))
        -V(8,8)*(V(9,1)*V(9,6)-V(9,4)*V(9,3))
        +V(8,9)*(V(9,1)*V(9,5)-V(9,2)*V(9,4));

    double a1 = V(9,1)*(V(8,5)*V(8,9)-V(8,6)*V(8,8))
        -V(9,2)*(V(8,4)*V(8,9)-V(8,7)*V(8,6))
        +V(9,3)*(V(8,4)*V(8,8)-V(8,5)*V(8,7))
        -V(9,4)*(V(8,2)*V(8,9)-V(8,3)*V(8,8))
        +V(9,5)*(V(8,1)*V(8,9)-V(8,3)*V(8,7))
        -V(9,6)*(V(8,1)*V(8,8)-V(8,2)*V(8,7))
        +V(9,7)*(V(8,2)*V(8,6)-V(8,3)*V(8,5))
        -V(9,8)*(V(8,1)*V(8,6)-V(8,4)*V(8,3))
        +V(9,9)*(V(8,1)*V(8,5)-V(8,2)*V(8,4));

    double a0 = V(8,1)*(V(8,5)*V(8,9)-V(8,8)*V(8,6))
        -V(8,2)*(V(8,4)*V(8,9)-V(8,7)*V(8,6))
        +V(8,3)*(V(8,4)*V(8,8)-V(8,7)*V(8,5));

    vector< double > *solutions = solve_cardano_real(a3,a2,a1,a0);

    // We are cheating here: Usually if there is more than
    // one solutions of lambda, one has to take all three
    // into account. This is not done here - only the
    // first solution is taken
    for(i=1;i<=9;i++)
        (*this)((i-1)/3+1,(i-1)%3+1)= V(8,i) + (*solutions)[0] * V(9,i);

    // Compute epipoles
    SVD(*this,D,U,V);

    m_leftepipole  = CPoint3D< double >(V.t().Row(3).t());
    m_rightepipole = CPoint3D< double >(U.t().Row(3).t());

    delete solutions;
}


/***************************************************************************/
/*!
 *  Eight point algorithm - suitable for eight and more points. This
 *  is based on Zisserman/Hartley page 263.
 *
 ***************************************************************************/
void CFundamentalMatrix::eightPointAlgorithm(const CFundamentalMatrix::TDataVector &list,
                                             unsigned int firstindex, unsigned int secondindex)
{

    Matrix A(max((int) list.size(),9),9);

    int i;
    CFundamentalMatrix::TDataVector::const_iterator currentpointpair;

    for(i=1, currentpointpair = list.begin(); currentpointpair != list.end() ;++i,++currentpointpair){
        A.Row(i) << (*currentpointpair)[firstindex].x * (*currentpointpair)[secondindex].x
            << (*currentpointpair)[firstindex].y * (*currentpointpair)[secondindex].x
            <<                                     (*currentpointpair)[secondindex].x
            << (*currentpointpair)[firstindex].x * (*currentpointpair)[secondindex].y
            << (*currentpointpair)[firstindex].y * (*currentpointpair)[secondindex].y
            <<                                     (*currentpointpair)[secondindex].y
            << (*currentpointpair)[firstindex].x
            << (*currentpointpair)[firstindex].y
            << 1;
    }

    // If underdetermined: Fill with zeros such that A is square
    if (list.size() == 8)
        A.Row(9) << 0 << 0 << 0 << 0 << 0 << 0 << 0 << 0 << 0;

    Matrix         V;
    Matrix         U;
    DiagonalMatrix D;

    SVD(A,D,U,V);

    V=V.t();

    // Fundamental Matrix without singularity constraint
    Matrix F(3,3);

    F << V(9,1) << V(9,2) << V(9,3)
        << V(9,4) << V(9,5) << V(9,6)
        << V(9,7) << V(9,8) << V(9,9);

    // Second step (see Zisserman 10.1.1 page 263: Enforcing singularity)

    SVD(F,D,U,V);

    // Elimination of the smallest singular value (t in Hartley/Zisserman)
    D(3,3) = 0;

    (*this) << U * D * V.t();

    m_leftepipole  = CPoint3D< double >(V.t().Row(3).t());
    m_rightepipole = CPoint3D< double >(U.t().Row(3).t());
}


/***************************************************************************/
/*!
 *  Helper function: Calculates one or three real roots of a polynomial
 *  having degree three.
 *
 ***************************************************************************/
vector< double > *CFundamentalMatrix::solve_cardano_real(const double &a,
                                                         const double &b,
                                                         const double &c,
                                                         const double &d)
{
    double q = 0.5*( (2*b*b*b)/(27*a*a*a) - (b*c)/(3*a*a) + d/a );
    double p = ( c - b*b/(3*a) ) / 3*a;

    complex<double> tmp = sqrt( complex<double> (q*q + p*p*p, 0.0) );
    complex<double> u   = pow ( complex<double> (-q,0.0) + tmp , 1.0/3.0 );
    complex<double> v   = pow ( complex<double> (-q,0.0) - tmp , 1.0/3.0 );


    complex<double> firstfactor (-sqrt(1.0/3.0)/2.0, sqrt(1.0/3.0)/2.0);
    complex<double> secondfactor(conj(firstfactor));

    complex<double> y[3];

    y[0] = u + v;
    y[1] = u*firstfactor +v*secondfactor;
    y[2] = u*secondfactor+v*firstfactor ;


    vector< double > *solutions = new vector< double >;

    for(int i=0;i<3;i++)
        if(y[i].imag() < 1E-5)
            solutions->push_back(y[i].real());

    return solutions;
}


/***************************************************************************/
/*!
 *  Calculates the distance of a point pair to the fundamental matrix
 *  based on geometric distances. This may be improved (see Hartley, page 269)
 *
 *  The distance is the one of one of the points to the epipolar line defined
 *  by the corresponding point. Taking the epipolar line defined by
 *  (a_x, a_y, a_z)' = F * x, then the distance to the euclidean point (b_x, b_y)
 *  is defined as | < b, a_(x,y) > + a_z | / ( | a_(x,y) | ) |
 *
 ***************************************************************************/
double CFundamentalMatrix::calculateDistance( const CFundamentalMatrix::TDataVector::value_type &sample,
                                             unsigned int firstindex, unsigned int secondindex)
{
    double distance = 0.0;
    double tmp;

    ColumnVector transformedpoint = *this * sample[firstindex].toColumnVector();

    tmp = ( sample[secondindex].x * transformedpoint(1) +
        sample[secondindex].y * transformedpoint(2) ) / sample[secondindex].z + transformedpoint(3);

    distance   = ( tmp*tmp ) / (transformedpoint(1)*transformedpoint(1) + transformedpoint(2)*transformedpoint(2));

    transformedpoint = this->t() * sample[secondindex].toColumnVector();

    tmp = ( sample[firstindex].x * transformedpoint(1) +
        sample[firstindex].y * transformedpoint(2) ) / sample[firstindex].z + transformedpoint(3);

    distance +=( tmp * tmp ) / (transformedpoint(1)*transformedpoint(1) + transformedpoint(2)*transformedpoint(2));

    return distance;
}
