/***************************************************************************
                          CTriangulation.cpp  -  description
                             -------------------
    begin                : 03.04.2003
    copyright            : (C) 2003 by Thomas Beilicke
    email                : tuvisionlib@tuvision.de
    origin               : Diplomarbeit Thomas Beilicke
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This file is part of the TUVision Library. It is free software; you   *
 *   can redistribute it and/or modify it under the terms of the           *
 *   GNU Library General Public License as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option)      *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

#include "tuv3d/CTriangulation.h"
#include "newmat/newmatap.h"	// For QR

using namespace TUVision;
/**********************************************************/ 
/*! The midpoint of the common perpendicular of the two
 *  rays is computed. The result is an euclidean point.
 *  Based on paper "Triangulation" by Hartley/Sturm, 1996.
 *  
 *  \remark This method nearly as fast as linearEigen but
 *          it gives considerably better results under the
 *			presence of noise.
 *
 *  \warning This method is neither projective nor affine
 *			invariant. It should only be used for
 *			euclidean reconstruction! 
 *
 *	\param P1 The first camera matrix.
 *	\param P2 The second camera matrix.
 *	\param corresp The correspondence on which the 
 *			triangulation is based. See camNo1 and camNo2. 
 *			The worldpoint set by this method is given in euclidean coordinates.
 *	\param camNo1 The number of the camera P1. A correspondence can 
 *			contain the image points of many cameras. The algorithm 
 *			will use the correspondences number camNo1 and camNo2 for 
 *			triangulation. The camera number starts with 0 as in 
 *			CWorldCorrespondence. The default is the camerapair 0 and 1.
 *	\param camNo2 Same as above for camera P2.
 */
void CTriangulation::midpoint(const CCameraMatrix &P1, const CCameraMatrix &P2, 
                              CCorrespondence &corresp, unsigned int camNo1, unsigned int camNo2)
{

    // Convert correspondence points to NEWMAT vectors
    NEWMAT::ColumnVector x1=corresp.getImagePoint(camNo1).toColumnVector();	// Make sure, the correspondence exists
    NEWMAT::ColumnVector x2=corresp.getImagePoint(camNo2).toColumnVector();	// Otherwise an exception is thrown by "vector"

    // decompose camera
    NEWMAT::Matrix M1=P1.SubMatrix(1,3,1,3).i();
    NEWMAT::Matrix M2=P2.SubMatrix(1,3,1,3).i();

    // Calculate camera centres	(don't use calcCentre because M is needed later anyway!)
    NEWMAT::ColumnVector p=P1.Column(4);
    NEWMAT::ColumnVector c1=-M1*p;
    p=P2.Column(4);
    NEWMAT::ColumnVector c2=-M2*p;

    // Build least squares equations A*x=b
    NEWMAT::Matrix A(3,2);
    A.Row(1) << (M1.Row(1)*x1).AsScalar() << (-M2.Row(1)*x2).AsScalar(); 
    A.Row(2) << (M1.Row(2)*x1).AsScalar() << (-M2.Row(2)*x2).AsScalar();
    A.Row(3) << (M1.Row(3)*x1).AsScalar() << (-M2.Row(3)*x2).AsScalar();

    NEWMAT::ColumnVector b=c2-c1;

    // QR decomposition to solve least squares
    NEWMAT::UpperTriangularMatrix U; NEWMAT::ColumnVector M;
    NEWMAT::QRZ(A,U);
    NEWMAT::QRZ(A,b,M);
    NEWMAT::ColumnVector alpha=U.i()*M;

    // Calculate midpoint out of result vector
    NEWMAT::ColumnVector result=(c1+alpha(1)*M1*x1+c2+alpha(2)*M2*x2)/2;

    // Update the correspondence's world point
    corresp.setEuclideanWorldPoint(result);
}

/**********************************************************/ 
/*! The linear-eigen method solves a system of equations
 *  by minimization (not iterative!). The result is a
 *  homogeneous point in general. Based on Hartley/Zisserman
 *  p. 297.
 *   
 *  \remark This is the fastest method but it gives the worst
 *			results for a noised euclidean reconstruction.
 *
 *  \warning This method is neither projective nor affine
 *			invariant. It should only be used for
 *			euclidean reconstruction! 
 *
 *	\param P1 The first camera matrix.
 *	\param P2 The second camera matrix.
 *	\param corresp The correspondence on which the 
 *			triangulation is based. See camNo1 parameter.
 *			The worldpoint set by this method is given in homogeneous coordinates.
 *	\param camNo1 The number of the camera P1. A correspondence can 
 *			contain the image points of many cameras. The algorithm 
 *			will use the correspondences number camNo1 and camNo2 for 
 *			triangulation. The camera number starts with 0 as in 
 *			CWorldCorrespondence. The default is the camerapair 0 and 1.
 *	\param camNo2 Same as above for camera P2.
 */
void CTriangulation::linearEigen(const CCameraMatrix &P1, const CCameraMatrix &P2, 
                                 CCorrespondence &corresp, unsigned int camNo1, unsigned int camNo2) 
{

    TUVision::CPoint3D<double> x1=corresp.getImagePoint(camNo1);	// Make sure, the correspondence exists
    TUVision::CPoint3D<double> x2=corresp.getImagePoint(camNo2);	// Otherwise an exception is thrown by "vector"

    NEWMAT::Matrix A(4,4);
    A.Row(1)	<< -x1.x*P1.Row(3)+x1.z*P1.Row(1);
    A.Row(2)	<<  x1.y*P1.Row(3)-x1.z*P1.Row(2);
    A.Row(3)	<< -x2.x*P2.Row(3)+x2.z*P2.Row(1);
    A.Row(4)	<<  x2.y*P2.Row(3)-x2.z*P2.Row(2);

    NEWMAT::Matrix U,V;
    NEWMAT::DiagonalMatrix D;

    NEWMAT::SVD(A,D,U,V);	

    TUVision::CPoint4D<double> result=(NEWMAT::ColumnVector) V.Column(4);
    corresp.setWorldPoint(result);
}

/**********************************************************/ 
/*! The iterative-eigen method iteratively uses linear-eigen.
 *  Weights are calculated to minimize an error that has
 *  a geometric meaning for the reconstruction.
 *  Based on paper "Triangulation" by Hartley/Sturm, 1996.
 *   
 *  \remark This method is slow but it gives by far the best
 *			results. The results even come close to the polynomial
 *			method (to be added later) which is optimal for 
 *			gaussian noise.
 *
 *  \warning This method is neither projective nor affine
 *			invariant. It should only be used for
 *			euclidean reconstruction! 
 *
 *	\param P1 The first camera matrix.
 *	\param P2 The second camera matrix.
 *	\param corresp The correspondence on which the 
 *			triangulation is based. See camNo1. The worldpoint set
 *			by this method is given in homogeneous coordinates.
 *	\param iterations The number of iterations. The default
 *		   is 15. In my tests more iterations have given no improvement.
 *	\param camNo1 The number of the camera P1. A correspondence can 
 *			contain the image points of many cameras. The algorithm 
 *			will use the correspondences number camNo1 and camNo2 for 
 *			triangulation. The camera number starts with 0 as in 
 *			CWorldCorrespondence. The default is the camerapair 0 and 1.
 *	\param camNo2 Same as above for camera P2.
 *	\returns The number of iterations. If this number matches #iterations,
 *			 probably no convergence was achieved. In this case
 *			 the polynomial method could be used afterwards.
 */
unsigned int CTriangulation::iterativeEigen(const CCameraMatrix &P1, const CCameraMatrix &P2, 
                                            CCorrespondence &corresp, unsigned int iterations, 
                                            unsigned int camNo1, unsigned int camNo2) 
{

    const double THRESH=1e-9;

    TUVision::CPoint3D<double> x1=corresp.getImagePoint(camNo1);	// Make sure, the correspondence exists
    TUVision::CPoint3D<double> x2=corresp.getImagePoint(camNo2);	// Otherwise an exception is thrown by "vector"

    // Normalize coordinates, saves computation for each loop
    x1/=x1.z;
    x2/=x2.z;

    // Initial weights
    double w1=1.0;   
    double w2=1.0;	

    NEWMAT::Matrix A(4,4);
    NEWMAT::ColumnVector result;
    double old_w1=0.0;
    double old_w2=0.0;

    NEWMAT::RowVector p1_1, p1_2, p1_3, p2_1, p2_2, p2_3;
    NEWMAT::Matrix U,V;
    NEWMAT::DiagonalMatrix D;

    unsigned int i;

    for (i=0; i< iterations; ++i) {
        p1_1=P1.Row(1);	// Slight speed increase if accessing only once
        p1_2=P1.Row(2);
        p1_3=P1.Row(3);
        p2_1=P2.Row(1);
        p2_2=P2.Row(2);
        p2_3=P2.Row(3);

        A.Row(1)	<< (-x1.x*p1_3+p1_1)/w1;
        A.Row(2)	<< ( x1.y*p1_3-p1_2)/w1;
        A.Row(3)	<< (-x2.x*p2_3+p2_1)/w2;
        A.Row(4)	<< ( x2.y*p2_3-p2_2)/w2;

        NEWMAT::SVD(A,D,U,V);	

        result=V.Column(4);

        old_w1=w1;	// store old weights
        old_w2=w2;
        w1=(p1_3*result).AsScalar();	// result is a columnvector of size 1
        w2=(p2_3*result).AsScalar();

        if (fabs(w1-old_w1)<THRESH && fabs(w2-old_w2)<THRESH)	// No more changes in weights => convergence
            break;  // stop if converging		
    }

    corresp.setWorldPoint(result);	
    return i;	// Return number of iterations
}

